/******************************************************************************
	disc: coulomb failure modeling 

	*******************************************************************
	* (c) Copyright 1994 The Regents of the University of California. *
	*     All Rights Reserved.                                        *
	*                                                                 *
	*     This work was produced under the sponsorship of the         *
	*     United States Department of Energy.                         *
	*     The Government retains certain rights therein.              *
	*******************************************************************

	usage:   disc < inputFile > failureFile

	compile: cc -g -o disc disc.c readinput.a -lm 

	disc calculates the 3-dimensional coulomb failure criteria 
   at specified points. points can be entered as longitude/latitude/depth 
   or x/y/z. input is modified version of disn stress output file.

	author: shawn larsen
	date:   march 1994
	where:  llnl
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "disl.h"
#include "usage.h"

#define AINC      10.00
#define DINC      10.00
#define VALUE     -5

main()
{
	double Cr, Cl, Cn, Ct;
	char line[256];
	stressTensor stress[1];

	/* print header */
	printf("LONGITUDE LATITUDE DEPTH   NAME      Cr      Cl      Cn      ");
	printf("Ct     MAX\n");
	printf("--------- --------  ----   ----     ---     ---     ---     ");
	printf("---     ---     ---\n");

	/* loop through all points printing maximum stress values */
	while(gets(line)) {
		if (getLine(line) == VALUE) {
			sscanf(line + 31, "%lf %lf %lf %lf %lf %lf", 
			       &stress->Txx, &stress->Tyy, &stress->Tzz, 
			       &stress->Txy, &stress->Txz, &stress->Tyz);
			getCoulomb(stress);
			line[31] = (char) '\0';
			printf("%31s", line);
			printf("%8.2lf%8.2lf%8.2lf%8.2lf%8.2lf    0.00\n", 
			       stress->Txx, stress->Tyy, stress->Tzz, stress->Txy, 
			       stress->Txz);
			}
		else {
			/*
			printf("s\n", line);
			*/
			}

		}
}




/******************************************************************************
   getCoulomb: get maxium failure criteria for at point

   usage: getCoulomb(stress)

	stress - input stress tensor

	Cr     - right-lateral failure
	Cl     - left -lateral failure
	Cn     - normal        failure
	Ct     - thrust        failure

   note: Cr > 0 is earthquake inducing (right-lateral)
   note: Cl < 0 is earthquake inducing (left- lateral)
   note: Cn > 0 is earthquake inducing (normal)
   note: Ct < 0 is earthquake inducing (thrust)
******************************************************************************/
getCoulomb(stress)
stressTensor *stress;
{
	double Nn[3], Ns[3], Nd[3], Ts, Td, Tn, Cr, Cl, Cn, Ct;
	double MaxCr, MaxCl, MaxCn, MaxCt, MaxCc, MaxCs, MaxCd;
	double angle, dip;
	int i, j;

	/* initialize */
	MaxCr = 0.;
	MaxCl = 0.;
	MaxCn = 0.;
	MaxCt = 0.;
	MaxCc = 0.;
	MaxCs = 0.;
	MaxCd = 0.;

	/* loop over all angles, dips */
	for(angle = 0.; angle < 360.; angle += AINC) {
		for(dip = 0.; dip <= 90.; dip += DINC) {
			getPlaneNormal(angle, dip, Nn, Ns, Nd);
			getPlaneStress(stress, Nn, Ns, Nd, &Ts, &Td, &Tn);
			Cr = Ts + FRICTION*Tn; /* > 0: earthquake inducing */
			Cl = Ts - FRICTION*Tn; /* < 0: earthquake inducing */
			Cn = Td + FRICTION*Tn; /* > 0: earthquake inducing */
			Ct = Td - FRICTION*Tn; /* < 0: earthquake inducing */
			if (Cr > MaxCr) MaxCr = Cr;
			if (Cl < MaxCl) MaxCl = Cl;
			if (Cn > MaxCn) MaxCn = Cn;
			if (Ct < MaxCt) MaxCt = Ct;
			}
		}

	/* get absolute maximums */
	if ( MaxCr > MaxCc) MaxCc =  MaxCr;
	if (-MaxCl > MaxCc) MaxCc = -MaxCl;
	if ( MaxCn > MaxCc) MaxCc =  MaxCn;
	if (-MaxCt > MaxCc) MaxCc = -MaxCt;
	if ( MaxCr > MaxCs) MaxCs =  MaxCr;
	if (-MaxCl > MaxCs) MaxCs = -MaxCl;
	if ( MaxCn > MaxCd) MaxCd =  MaxCn;
	if (-MaxCt > MaxCd) MaxCd = -MaxCt;

	/* set to output */
	stress->Txx = MaxCr;
	stress->Tyy = MaxCl;
	stress->Tzz = MaxCn;
	stress->Txy = MaxCt;
	stress->Txz = MaxCs;

}




/******************************************************************************
	getPlaneNormal: determine normal vectors to plane

	usage: getPlaneNormal(azimuth, dip, Nn, Ns, Nd)

	azimuth - strike of plane orientation (clockwise from east)
	dip     - to north for 0 azimuth (to south for 180 azimuth)
	Nn      - normal to plane (outward if dipping toward you)
	Ns      - normal along strike (goes to left)
	Nd      - normal along dip (down if dipping toward you)

	note: normal vectors defined where N[0]=east, N[1]=north, N[2]=up
******************************************************************************/
getPlaneNormal(azimuth, dip, Nn, Ns, Nd)
double azimuth, dip, *Nn, *Ns, *Nd;
{
	azimuth =        azimuth/(RAD);
	dip     =            dip/(RAD);
	Nn[0] = -sin(dip)*sin(azimuth);
	Nn[1] =  sin(dip)*cos(azimuth);
	Nn[2] =  cos(dip)             ;
	Ns[0] =           cos(azimuth);
	Ns[1] =           sin(azimuth);
	Ns[2] =                     0.;
	Nd[0] = -cos(dip)*sin(azimuth);
	Nd[1] =  cos(dip)*sin(azimuth);
	Nd[2] = -sin(dip)             ;
}





/******************************************************************************
	getPlaneStress: use stress tensor to get stresses at specified orientation

	usage: getPlaneStress(stress, Nn, Ns, Nd, &sshear, &dshear, &normal)

	stress  - stress tensor information (or strain)
	Nn      - normal to plane (outward if dipping toward you)
	Ns      - normal along strike (goes to left)
	Nd      - normal along dip (down if dipping toward you)
	sshear  - (return) stike shear (right-lateral is positive)
	dshear  - (return) dipping shear (down-dip is positive) 
	normal  - (return) normal stress (tension is positive)

	note: normal vectors defined where N[0]=east, N[1]=north, N[2]=up
******************************************************************************/
getPlaneStress(stress, Nn, Ns, Nd, sshear, dshear, normal)
stressTensor *stress;
double *Nn, *Ns, *Nd, *sshear, *dshear, *normal;
{
	double T[3]; 

	/* determine traction vector */
	T[0] = stress->Txx*Nn[0] + stress->Txy*Nn[1] + stress->Txz*Nn[2];
	T[1] = stress->Txy*Nn[0] + stress->Tyy*Nn[1] + stress->Tyz*Nn[2];
	T[2] = stress->Txz*Nn[0] + stress->Tyz*Nn[1] + stress->Tzz*Nn[2];

	/* determine fault stresses */
	*normal = T[0]*Nn[0] + T[1]*Nn[1] + T[2]*Nn[2];
	*sshear = T[0]*Ns[0] + T[1]*Ns[1] + T[2]*Ns[2];
	*dshear = T[0]*Nd[0] + T[1]*Nd[1] + T[2]*Nd[2];
	
}
