sign convention in getslip (right vs left, etc)
/******************************************************************************
	disf: 3-D fault failure sequence by dislocations

	*******************************************************************
	* (c) Copyright 1992 The Regents of the University of California. *
	*     All Rights Reserved.                                        *
	*                                                                 *
	*     This work was produced under the sponsorship of the         *
	*     United States Department of Energy.                         *
	*     The Government retains certain rights therein.              *
	*******************************************************************

	usage:   disf [optional parameters] < inputFile > outputFile

	compile: cc -g -o disf disf.c getdisl.c readinput.a -lm -lget

	disf determines the stress induced failure sequence for 3-D fault geometries 
	using the dislocation models of chinnery, BSSA 1961 (#3), 
	savage and hastie, JGR 1966 (#20), mansinha and smylie, JGR 1967 (#18), 
	and mansinha and smylie, BSSA 1971 (#5).
	a file containing the name of all paramaters, faults, and 
	locations is used as input.
	points can be entered as longitude/latitude/depth or x/y/z.

	author: shawn larsen
	date:   october 1992
	where:  llnl
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "disl.h"
#include "usage.h"

struct data_group {
	faultType       *faults;
	double         **sshear;
	double         **dshear;
	double         **normal;
	double            *slip;
	double          *stress;
	char        **fileLines;
	};

typedef struct data_group Data;
typedef struct parameter_faultfailure parameters;

main(argc, argv)
int    argc;
char **argv;
{
	Data          data[1];
	parameters     par[1];
	int i;

	
	/* check for no input (print usage message) */
	if (argc == 1) if (checkUsage()) exit(0);

	
	/* get initial parameters */
	setpar(argc, argv);
	getPar(par);
	endpar();


	/* read input file */
	readInput(par, data);


	/* get input parameters and data */
	getInput(par, data);


	/* get failure stress kernal */
	stressKernal(par, data);


	/* run failure loop */
	failureLoop(par, data);


	/* write output (stress and strain point to same) */
	writeOutput(par, data);


	/* free memory */
	freeMemory(par, data);

}




/******************************************************************************
	checkUsage: check usage and print usage message if no input

	usage: checkUsage()

	return 0: correct usage - continue execution
	       1: print   usage - stop     execution
******************************************************************************/
checkUsage()
{
	struct stat statbuf;
	int i;
	double increment;
	int bmax, fmax, m, xy, strain;

	/* get input status */
	fstat(0, &statbuf);

	/* return if normal input */
	if (statbuf.st_mode & REGULAR_FILE) return(0);
	if (statbuf.st_mode & PIPE)         return(0);

	/* print usage and return exit */
	i = 0;
	while(disfUsage[i]) fprintf(stderr, "%s", disfUsage[i++]);
	return(1);

}




/******************************************************************************
	freeMemory: free memory

	usage: freeMemory(par, data)

	par  - parameters
	data - input data
******************************************************************************/
freeMemory(par, data)
parameters  *par;
Data       *data;
{
	int i;
	free(data->faults);
	for(i = 0; i < par->f; i++) {
		free(sshear[i]);
		free(dshear[i]);
		free(normal[i]);
		}
	free(sshear);
	free(dshear);
	free(normal);
	free(stress);
	free(slip);
	for(i = 0; i < par->n; i++) free(data->fileLines[i])  ;free(data->fileLines);
}




/******************************************************************************
	getPar: read program parameters

	usage: getPar(par)

	par - parameters
******************************************************************************/
getPar(par)
parameters *par;
{
	int i;

	/* initialize parameters */
	par->f            = 0;
	par->XY           = XYFLAG;
	par->increment    = INCREMENT;
	par->friction     = FRICTION;
	par->lamda        = LAMDA;
	par->mu           = MU;
	par->t            = TIMESTEPS;
	par->dt           = TIMEINCREMENT;
	par->sstrength    = SSTRENGTH;
	par->dstrength    = DSTRENGTH;
	par->astress      = ASTRESS;

	/* get parameters */
	getpar("xy"          , "d", &par->XY); 
	getpar("t"           , "d", &par->t); 
	getpar("increment"   , "F", &par->increment);
	getpar("dt"          , "F", &par->tincrement);
	getpar("sstrength"   , "F", &par->sstrength);
	getpar("dstrength"   , "F", &par->dstrength);
	getpar("astress"     , "F", &par->astress);

	/* check parameters */
	if (par->increment <=        0.) par->increment = INCREMENT;
	if (par->t         <         0 ) par->t         = TIMESTEPS;
	if (par->dt        <=        0.) par->dt        = TINCREMENT;
	if (par->sstrength <=        0.) par->sstrength = SSTRENGTH;
	if (par->dstrength <=        0.) par->dstrength = DSTRENGTH;
	if (par->astress   <=        0.) par->astress   = ASTRESS;  

	/* check if parameter match from readline */
	if (getType("COMMENT") - LINECOMMENT) 
		fprintf(stderr, "WARNING: readline mismatch - LINECOMMENT\n");
	if (getType("NULL")    - LINENULL)    
		fprintf(stderr, "WARNING: readline mismatch - LINENULL\n");
	if (getType("BLANK") - LINEBLANK) 
		fprintf(stderr, "WARNING: readline mismatch - LINEBLANK\n");
	if (getType("VALUE") - LINEVALUE) 
		fprintf(stderr, "WARNING: readline mismatch - LINEVALUE\n");
	if (getType("PAR") - LINEPAR) 
		fprintf(stderr, "WARNING: readline mismatch - LINEPAR\n");
	if (getType("UNKNOWN") - LINEUNKNOWN) 
		fprintf(stderr, "WARNING: readline mismatch - LINEUNKNOWN\n");

}




/******************************************************************************
	readInput: read input files and store in argv array

	usage: readInput(par, data)

	par    - parameters
	data   - information
******************************************************************************/
readInput(par, data)
parameters *par;
Data       *data;
{
	int n, maxLines;
	unsigned size;

	/* initialize */
	maxLines = FILELINES;
	n = 0;

	/* read input file and store */
	size = maxLines*sizeof(char *);
	data->fileLines = (char **) malloc(size);
	while((n += argvFile(data->fileLines + n, FILELINES, stdin)) >= maxLines) {
		maxLines += FILELINES;
		size = (unsigned) maxLines*sizeof(char *);
		data->fileLines = (char **) realloc((char *) data->fileLines, size);
		};
	par->n = n;

	/* null last pointer */
	data->fileLines[par->n] = NULL;

}




/******************************************************************************
	getInput: get input parameters and data (allocate memory)

	usage: getInput(par, data)

	par    - parameters
	data   - information
******************************************************************************/
getInput(par, data)
parameters  *par;
Data       *data;
{
	int lineType, b, f, g;
	char **line;

	/* set line types */
	setLine("fault ",      LINEFAULT);
   setLine("datum ",      LINEDATUM);
	setLine("options ",    LINEOPTION);
	setLine("stop ",       LINESTOP);

	/* count number of data items */
	par->f = 0;
	line = data->fileLines;
	while(*line != NULL) {
		switch((lineType = getLine(*line))) {
			case LINEFAULT:
				par->f++;
				break;
			default:
				break;
			}
		line++;
		}

	/* get memory */
	data->faults = (faultType    *) malloc(par->f*sizeof(faultType));

	/* read in faults and point groups */
	f = 0; 
	line = data->fileLines;
	while(*line != NULL) {
		switch(getLine(*line)) {
			case LINEFAULT:
				f += getFault(data->faults + f, *line);
				break;
			case LINEOPTION:
			case LINEPAR:
				setArg(*line);
				getOption(par, *line);
				endArg();
				break;
			case LINECOMMENT:
			case LINENULL:
			case LINEBLANK:
			case LINEVALUE:
			case LINEUNKNOWN:
			default:
				break;
			}
		line++;
		}

	/* end line types */
	endLine();

}




/******************************************************************************
	getFault: get fault

	usage: f = getFault(fault, line)

	fault  - fault information
	line   - input line

	return: f - number of faults (1)
******************************************************************************/
getFault(fault, line)
faultType *fault;   
char      *line;
{
   int input;
   char lineType[20], fType[20];
   input = sscanf(line, "%s %s %lf %lf %lf %lf %lf %lf %lf %lf %s %d %d",
                         lineType,     fType,
                        &fault->ds,   &fault->dl,   &fault->dt,   &fault->dw,
                        &fault->dp,   &fault->xLon, &fault->yLat, &fault->az,
                         fault->name, &fault->nx,   &fault->ny);
   fault->ft = *fType;
   if (input == 9) {
      fault->nx = 1;
      fault->ny = 1;
      }
   else if (input != 11) {
      fprintf(stderr, "ERROR: bad fault format\n");
      fprintf(stderr, "\"%s\"\n", line);
      exit(-1);
      }
	return(1);
}




/******************************************************************************
	getOption: get options

	usage: getOption(par, line)

	par    - parameters
	line   - input line
******************************************************************************/
getOption(par, line)
parameters  *par;
char       *line;
{
	getArg("xy"    ,    "d", &par->XY); 
	getArg("increment", "F", &par->increment);
	getArg("t",         "d", &par->t);
	getArg("dt",        "F", &par->dt);
}




/******************************************************************************
	getStressKernal: get dislocation stress kernal (coulomb stress)

	usage: getStressKernal(par, data)

	par    - parameters
	data   - data and information
******************************************************************************/
getStressKernal(par, data)
parameters     *par;
Data          *data;
{
	strainTensor strain[1], *stress;
	faultTypes   *faults;
	double **A, **B, **C, azimuth, dip;
	int i, j;
	faults = data->faults;
	A = data->sshear;
	B = data->dshear;
	C = data->normal;
	for(i = 0; i < par->f; i++) {
		getFaultPoints(strain, faults + i, par->XY);
		azimuth = faults[i].azimuth;
		dip     = faults[i].dip;
		for(j = 0; j < par->f; j++) {
			getStrain(par, faults + j, strain);
			stress = strain;
			getStress(par, stress);
			getFaultStress(stress, azimuth, dip);
			A[i][j] = stress->Txx;
			B[i][j] = stress->Tyy;
			C[i][j] = stress->Tzz;
			}
		}


}




getFaultPoints(strain, fault, XY)
strainTensor *strain;
faultType     *fault;
int               XY;
{
	double dp, nx, ny, x0, y0, z0, sdx, sdy, sdz, ddx, ddy, ddz, zdx, zdy;
	int i, j, n;

	/* initialize */
	n = 0;
	dp = fault->dp/(RAD);
	nx = (double) fault->nx;
	ny = (double) fault->ny;

	/* get subfault increments due to strike */
	sdx = fault->dl*cos(fault->az/(RAD))/nx;
	sdy = fault->dl*sin(fault->az/(RAD))/nx;

	/* get subfault increments due to dip */
	ddx = fault->dw*cos(dp)*cos((fault->az + 90.)/(RAD))/ny;
	ddy = fault->dw*cos(dp)*sin((fault->az + 90.)/(RAD))/ny;
	ddz = fault->dw*sin(dp)/ny;

	/* get initial increment due to fault depth (not for 0 dip) */
	zdx = fault->dt*cos(dp)/sin(dp)*cos((fault->az + 90.)/(RAD));
	zdy = fault->dt*cos(dp)/sin(dp)*sin((fault->az + 90.)/(RAD));

	/* convert into units of latitude and longitude */
	if (!XY) {
		sdx = sdx/(KMD)/cos(fault->yLat/(RAD));
		sdy = sdy/(KMD);
		ddx = ddx/(KMD)/cos(fault->yLat/(RAD));
		ddy = ddy/(KMD);
		zdx = zdx/(KMD)/cos(fault->yLat/(RAD));
		zdy = zdy/(KMD);
		}

	/* calculate starting point (lower-left) */
	x0 = fault->xLon + zdx + (ny - 0.5)*ddx - (nx/2. - 0.5)*sdx;
	y0 = fault->yLat + zdy + (ny - 0.5)*ddy - (nx/2. - 0.5)*sdy;
	z0 = fault->dt +         (ny - 0.5)*ddz;

	/* generate plane points row-by-row from lower-left */
   for(j = 0; j < fault->ny; j++) {
      for(i = 0; i < fault->nx; i++) {
         strain[n].xLon = x0 + ((double) i)*sdx - ((double) j)*ddx;
         strain[n].yLat = y0 + ((double) i)*sdy - ((double) j)*ddy;
         strain[n].zEle = z0                    - ((double) j)*ddz;
         n++;
         }
      }

}




/******************************************************************************
	getStrain: get strain tensor

	usage: getStrain(par, fault, strain)

	par    - parameters
	fault  - fault parameters
	strain - strain information
******************************************************************************/
getStrain(par, fault, strain)
parameters   *par;
faultType    *faults;
strainTensor *strain;
{
	benchMark point[6];
	double Ux, Uy, Uz;
	int i, j, k, XY;

	/* initialize */
	XY = par->XY;

	/* get 6 points for numerical computation */
	getPoints(point, strain, par->increment, par->XY);
	for(i = 0; i < 6; i++) {
		getDisl(&Ux, &Uy, &Uz, &(point[i]), fault, par->XY);
		point[i].xMod = Ux;
		point[i].yMod = Uy;
		point[i].zMod = Uz;
		}
	getStrainTensor(strain, point, par->increment);

}




getPoints(point, strain, increment, XY)
benchMark    *point;
strainTensor *strain;
double        increment;
int           XY;
{
	
	double dx, dy, dz;
	int i;

	/* get increments */
	dx = increment; dy = increment; dz = increment;
	if (!XY) {
		dx = increment/((KMD)*cos(strain->yLat/(RAD)));
		dy = increment/(KMD);
		}

	/* get initial points */
	for(i = 0; i < 6; i++) {
		point[i].xLon = strain->xLon;
		point[i].yLat = strain->yLat;
		point[i].zEle = strain->zEle;
		point[i].xMod = 0.;
		point[i].yMod = 0.;
		point[i].zMod = 0.;
		}

	/* get new points */
	point[0].xLon += dx;
	point[1].xLon -= dx;
	point[2].yLat += dy;
	point[3].yLat -= dy;
	point[4].zEle += dz;
	point[5].zEle -= dz;
	if (point[5].zEle < 0.) point[5].zEle = 0.;

}




getDisl(ux, uy, uz, point, f, XY)
benchMark *point;
faultType *f;
double *ux, *uy, *uz;
int XY;
{
	double fx, fy, fz, x, y, z, cs, sn;

	/* initialize */
	cs = cos(f->az/(RAD));
	sn = sin(f->az/(RAD));

	/* change from Cartesian to fault coordinates */
	fx = point->xLon - f->xLon;
	fy = point->yLat - f->yLat;
	if (!XY) {
		fx = fx*(KMD)*cos(f->yLat/(RAD));
		fy = fy*(KMD);
		}
	x = cs*fx + sn*fy;
	y = cs*fy - sn*fx;
	z = fabs(point->zEle);

	/* get displacements in fault coordinate system */
	getdisl(x, y, z, f->dl, f->dt, f->dw, f->dp, f->ft, &fx, &fy, &fz);

	/* rotate displacements from fault to Cartesian coordinates */
	cs = cos(f->az/(RAD));
	sn = sin(f->az/(RAD));
	*ux = fx*cs - fy*sn;
	*uy = fx*sn + fy*cs;
	*uz = fz;
	
}




getStrainTensor(strain, point, increment)
strainTensor *strain;
benchMark    *point;
double        increment;
{
	double dx, dy, dz;

	/* get increments */
	dx = 2.*increment;
	dy = 2.*increment;
	dz = 2.*increment;
	if ((fabs(point->zEle)-increment) < 0.) dz = increment + fabs(point->zEle); 

	/* change from km to cm */
	dx *= CMKM;
	dy *= CMKM;
	dz *= CMKM;

	/* get compressional strains (Txx, Tyy, Tzz) */
	strain->Txx = (point[0].xMod - point[1].xMod)/dx;
	strain->Tyy = (point[2].yMod - point[3].yMod)/dy;
	strain->Tzz = (point[5].zMod - point[4].zMod)/dz;

	/* get shear strains */ 
	strain->Txy = 0.5*((point[0].yMod - point[1].yMod)/dx + 
	                   (point[2].xMod - point[3].xMod)/dy);
	strain->Txz = 0.5*((point[0].zMod - point[1].zMod)/dx + 
	                   (point[5].xMod - point[4].xMod)/dz);
	strain->Tyz = 0.5*((point[2].zMod - point[3].zMod)/dy + 
	                   (point[5].yMod - point[4].yMod)/dz);

}





/******************************************************************************
	getStress: convert strain tensor to stress tensor

	usage: getStress(par, stress)

	par    - parameters
	stress - stress information
******************************************************************************/
getStress(par, stress)
parameters   *par;
stressTensor *stress;
{
	strainTensor *strain;
	double lamda, mu2, dilatation;
	int b;

	/* initialize */
	strain =    stress;
	lamda  =    par->lamda;
	mu2    = 2.*par->mu;

	/* get stress tensor for all points */
	for(b = 0; b < par->b; b++) {
		dilatation = lamda*(strain[b].Txx + strain[b].Tyy + strain[b].Tzz);
		stress[b].Txx = dilatation + mu2*strain[b].Txx;
		stress[b].Tyy = dilatation + mu2*strain[b].Tyy;
		stress[b].Tzz = dilatation + mu2*strain[b].Tzz;
		stress[b].Txy = mu2*strain[b].Txy;
		stress[b].Txz = mu2*strain[b].Txz;
		stress[b].Tyz = mu2*strain[b].Tyz;
		}

}




/******************************************************************************
	getPlaneNormal: determine normal vectors to plane

	usage: getPlaneNormal(azimuth, dip, Nn, Ns, Nd)

	azimuth - strike of plane orientation (clockwise from east)
	dip     - to north for 0 azimuth (to south for 180 azimuth)
	Nn      - normal to plane (outward if dipping toward you)
	Ns      - normal along strike (goes to left)
	Nd      - normal along dip (down if dipping toward you)

	note: normal vectors defined where N[0]=east, N[1]=north, N[2]=up
******************************************************************************/
getPlaneNormal(azimuth, dip, Nn, Ns, Nd)
double azimuth, dip, *Nn, *Ns, *Nd;
{
	azimuth =        azimuth/(RAD);
	dip     =            dip/(RAD);
	Nn[0] = -sin(dip)*sin(azimuth);
	Nn[1] =  sin(dip)*cos(azimuth);
	Nn[2] =  cos(dip)             ;
	Ns[0] =           cos(azimuth);
	Ns[1] =           sin(azimuth);
	Ns[2] =                     0.;
	Nd[0] = -cos(dip)*sin(azimuth);
	Nd[1] =  cos(dip)*sin(azimuth);
	Nd[2] = -sin(dip)             ;
}





/******************************************************************************
	getPlaneStress: use stress tensor to get stresses at specified orientation

	usage: getPlaneStress(stress, Nn, Ns, Nd, &sshear, &dshear, &normal)

	stress  - stress tensor information (or strain)
	Nn      - normal to plane (outward if dipping toward you)
	Ns      - normal along strike (goes to left)
	Nd      - normal along dip (down if dipping toward you)
	sshear  - (return) stike shear (right-lateral is positive)
	dshear  - (return) dipping shear (down-dip is positive) 
	normal  - (return) normal stress (tension is positive)

	note: normal vectors defined where N[0]=east, N[1]=north, N[2]=up
******************************************************************************/
getPlaneStress(stress, Nn, Ns, Nd, sshear, dshear, normal)
stressTensor *stress;
double *Nn, *Ns, *Nd, *sshear, *dshear, *normal;
{
	double T[3]; 

	/* determine traction vector */
	T[0] = stress->Txx*Nn[0] + stress->Txy*Nn[1] + stress->Txz*Nn[2];
	T[1] = stress->Txy*Nn[0] + stress->Tyy*Nn[1] + stress->Tyz*Nn[2];
	T[2] = stress->Txz*Nn[0] + stress->Tyz*Nn[1] + stress->Tzz*Nn[2];

	/* determine fault stresses */
	*normal = T[0]*Nn[0] + T[1]*Nn[1] + T[2]*Nn[2];
	*sshear = T[0]*Ns[0] + T[1]*Ns[1] + T[2]*Ns[2];
	*dshear = T[0]*Nd[0] + T[1]*Nd[1] + T[2]*Nd[2];
	
}




/******************************************************************************
	failureLoop: loop over failure criteria for faults

	usage: failureLoop(par, data)

	par  - parameters 
	data - information
******************************************************************************/
failureLoop(par, data)
parameters  *par;
Data       *data;
{
	int i, t;
	double *slip, *stress;
	int    *failMask;

	/* initialize and allocate temporary arrays */
	failMask = (char   *) malloc(par->f*sizeof(char  ));
	slip     = (double *) malloc(par->f*sizeof(double));
	nullVectors(par, slip, failMask);
	stress = data->stress;

	/* loop over all time steps */
	for(t = 0; t < par->t; t++) {
		driveStress(par, stress);
		while(getSlip(par, stress, slip, failMask, data->sshear)) {
			updateStress(par, stress, slip, failMask, data->sshear);
			}
		updateInfo(par, slip, data);
		nullVectors(par, slip, failMask);
		}

	/* free temporary memory */
	free(failMask);
	free(slip);

}




nullVectors(par, slip, failMask)
parameters      *par;
double         *slip;
int        *failMask;
{
	int i;
	for(i = 0; i < par->f; i++) {
		slip[i]     =          0.;
		failMask[i] = (char) '\0';
		}
}




/******************************************************************************
	driveStress: determine stress update from tectonic loading

	usage: driveStress(par, stress)

	par      - parameters
	stress   - stress level on fault
******************************************************************************/
driveStress(par, stress)
parameters   *par;
double       *stress;
{

}




/******************************************************************************
	getSlip: get slip on subfaults > critical stress 

	usage: getSlip(par, slip, stress, failMask, A)

	par      - parameters
	slip     - slip on fault this iteration
	stress   - stress level on fault
	failMask - failure mode of fault (0=none; 1=static; 2=dynamic)
	A        - stress kernal
******************************************************************************/
getSlip(par, slip, stress, failMask, sshear)
parameters   *par;
double       *slip, *stress, **sshear;
int          *failMask;
{
	int i, failure = 0;
	for(i = 0; i < par->f; i++) {
		if      (stress[i] >= par->sstrength) {
			slip[i]     = (par->astress - stress[i])/sshear[i][i];
			failMask[i] = 2;
			failure++;
			}
		else if (stress[i] >= par->dstrength && failMask[i]) {
			slip[i]     = (par->astress - stress[i])/sshear[i][i];
			failMask[i] = 2;
			failure++;
			}
		else {
			if (failMask[i]) failMask[i] = 1;
			}
		}
	return(failure);
}

check signs


/******************************************************************************
	updateStress: get new stress on all faults based on previous fault slips 

	usage: updateStress(par, slip, stress, failMask)

	par      - parameters
	slip     - slip on fault this iteration
	stress   - stress level on fault
	failMask - failure mode of fault (0=none; 1=static; 2=dynamic)
******************************************************************************/
updateStress(par, slip, stress, failMask, sshear)
parameters   *par;
double       *slip, *stress, **sshear;
int          *failMask;
{
	int i, j;
	for(i = 0; i < par->f; i++) if (failMask[i] == 2)
			for(j = 0; j < par->f; j++) stress[j] += sshear[j][i]*slip[i];
}




/******************************************************************************
	updateInfo: update parameters and information after this time step

	usage: updateInfo(par, slip, data)

	par  - parameters
	slip - slip on fault this iteration
	data - information 
******************************************************************************/
updateInfo(par, slip, data)
parameters   *par;
double       *slip;
Data         *data;
{
	int i;
	for(i = 0; i < ; i++) {
		data->slip[i] += slip[i];
		}
}




/******************************************************************************
	writeOutput: write results to standard output

	usage: writeOutput(par, data)

	par    - parameters
	data   - input data
******************************************************************************/
writeOutput(par, data)
parameters   *par;
Data        *data;
{
	strainTensor *strain;
	stressTensor *stress;
	PointGroup   *groups;
	faultType         *f;
	int          i, j, n;
	char      format[82];

	/* initialize */
	groups = data->groups;
	f      = data->faults;

   /* write header and input parameters */
   printf("                              STRESS/STRAIN MODELING\n");
   printf("                              ----------------------\n");
   printf("options xy=%-d stress=%-d mdigit=%-d increment=%-8.5lf\n", 
			par->XY, par->stress, par->mDigit, par->increment);

	/* write header information for faults */
	printFLabel(par->XY);

	/* write faults */
   for(i = 0; i < par->f; i++) {
      printf("fault ");
      printf("%c %8.2lf %7.2lf %6.2lf %6.2lf %5.1lf %9.4lf %8.4lf %6.1lf  ",
              f[i].ft, f[i].ds, f[i].dl, f[i].dt, f[i].dw, f[i].dp,
              f[i].xLon, f[i].yLat, f[i].az);
      printf("%-7s",  f[i].name);
      printf("\n");
      }

	/* write header information for strain/stress tensor */
	printf("\n");
	printf("\n");
	if (!par->stress) {
		printf("                                 MICROSTRAIN TENSOR\n");
		printf("                                 ------------------\n");
		}
	else {
		printf("                                STRESS TENSOR (BARS)\n");
		printf("                                -------------------\n");
		}

	/* get default format string */
   sprintf(format, "%%9. lf%%8. lf%%7. lf%%+7s%%8. lf%%8. lf%%8. lf");
   sprintf(format + 40, "%%8. lf%%8. lf%%8. lf\n\0");

	/* write stress/strain tensor for points */
	n = 0;
	for(i = 0; i < par->g; i++) {
		if (groups[i].type == LINEPOINT) {
			if (n++ == 0) printSLegend(par->XY, NULL,  LINEPOINT);
			stress = groups[i].tensor;
			getFormat(format, par, stress + 0);
			printf(format, stress[0].xLon, stress[0].yLat, stress[0].zEle, 
			               stress[0].name,
			               stress[0].Txx , stress[0].Tyy , stress[0].Tzz ,
			               stress[0].Txy , stress[0].Txz , stress[0].Tyz );
			}
		}

	/* write stress/strain tensor for datum */
	for(i = 0; i < par->g; i++) {
		if (groups[i].type == LINEDATUM) {
			printSLegend(par->XY, groups[i].line, LINEDATUM);
			stress = groups[i].tensor;
			n = groups[i].nx*groups[i].ny;
			for(j = 0; j < n; j++) {
				getFormat(format, par, stress + j);
				printf(format, stress[j].xLon, stress[j].yLat, stress[j].zEle, 
				               stress[j].name,
				               stress[j].Txx , stress[j].Tyy , stress[j].Tzz ,
				               stress[j].Txy , stress[j].Txz , stress[j].Tyz );
				}
			}
		}

	/* write stress/strain tensor for datum */
	for(i = 0; i < par->g; i++) {
		if (groups[i].type == LINEBLOCK) {
			printSLegend(par->XY, groups[i].line, LINEBLOCK);
			stress = groups[i].tensor;
			n = groups[i].nx*groups[i].ny*groups[i].nz;
			for(j = 0; j < n; j++) {
				getFormat(format, par, stress + j);
				printf(format, stress[j].xLon, stress[j].yLat, stress[j].zEle, 
				               stress[j].name,
				               stress[j].Txx , stress[j].Tyy , stress[j].Tzz ,
				               stress[j].Txy , stress[j].Txz , stress[j].Tyz );
				}
			}
		}

	/* write stress/strain tensor for datum */
	for(i = 0; i < par->g; i++) {
		if (groups[i].type == LINEPLANE) {
			printSLegend(par->XY, groups[i].line, LINEPLANE);
			stress = groups[i].tensor;
			n = groups[i].nx*groups[i].ny;
			for(j = 0; j < n; j++) {
				getFormat(format, par, stress + j);
				printf(format, stress[j].xLon, stress[j].yLat, stress[j].zEle, 
				               stress[j].name,
				               stress[j].Txx , stress[j].Tyy , stress[j].Tzz ,
				               stress[j].Txy , stress[j].Txz , stress[j].Tyz );
				}
			}
		}

}




printFLabel(XY) 
int XY;
{
	printf("\n");
	printf("\n");
	printf("                                  FAULTS MODELED\n");
	printf("                                  --------------\n");
	printf("\n");
   if (XY) {
      printf("    Ftype   Slip  Length  Depth  Width   Dip ");
      printf("    XDist    YDist Azimuth Name\n");
      printf("    -----   ----  ------  -----  -----   --- ");
      printf("    -----    ----- ------- ----\n");
      }
   else {
      printf("    Ftype  Slip   Length  Depth  Width   Dip ");
      printf("Longitude Latitude Azimuth Name\n");
      printf("    -----  ----   ------  -----  -----   --- ");
      printf("--------- -------- ------- ----\n");
      }
}




printSLegend(XY, line, mode)
int XY, mode;
char *line;
{
   int n = 0;
   char string[258];
   switch(mode) {
      case LINEPOINT:
         printf("\n");
         printf("POINT\n");
         printf("\n");
         printSLabel(XY);
         break;
		case LINEDATUM:
         while(line[n] == (char) ' ') n++;
         printf("\n\n");
         strcpy(string, line + n);
         strncpy(string, "DATUM", 5);
         printf("%s\n", string);
         printf("\n");
         printSLabel(XY);
         break;
		case LINEBLOCK:
         while(line[n] == (char) ' ') n++;
         printf("\n\n");
         strcpy(string, line + n);
         strncpy(string, "BLOCK", 5);
         printf("%s\n", string);
         printf("\n");
         printSLabel(XY);
         break;
		case LINEPLANE:
         while(line[n] == (char) ' ') n++;
         printf("\n\n");
         strcpy(string, line + n);
         strncpy(string, "PLANE", 5);
         printf("%s\n", string);
         printf("\n");
         printSLabel(XY);
         break;
      }
}




printSLabel(XY)
int XY;
{
   if (XY) {
      printf("    XDIST   YDIST  DEPTH   NAME     Txx     Tyy     Tzz    ");
      printf(" Txy     Txz     Tyz\n");
      printf("    -----   -----  -----   ----     ---     ---     ---    ");
      printf(" ---     ---     ---\n");
      }
   else {
      printf("LONGITUDE LATITUDE  ELEV   NAME     Txx     Tyy     Tzz    ");
      printf(" Txy     Txz     Tyz\n");
      printf("--------- --------  ----   ----     ---     ---     ---    ");
      printf(" ---     ---     ---\n");
      }
}




getFormat(format, par, stress)
parameters   *par;
stressTensor *stress;
char         *format;
{
	format[ 3] = (char) (getMDigit(stress->xLon, 9,           4) + '0');
	format[ 9] = (char) (getMDigit(stress->yLat, 8,           4) + '0');
	format[15] = (char) (getMDigit(stress->zEle, 7,           2) + '0');
	format[25] = (char) (getMDigit(stress->Txx,  8, par->mDigit) + '0');
	format[31] = (char) (getMDigit(stress->Tyy,  8, par->mDigit) + '0');
	format[37] = (char) (getMDigit(stress->Tzz,  8, par->mDigit) + '0');
	format[43] = (char) (getMDigit(stress->Txy,  8, par->mDigit) + '0');
	format[49] = (char) (getMDigit(stress->Txz,  8, par->mDigit) + '0');
	format[55] = (char) (getMDigit(stress->Tyz,  8, par->mDigit) + '0');
}




int
getMDigit(value, field, mDigit)
double value;
int    field, mDigit;
{
	int max, n;
	char fTest[20], mTest[20];
	sprintf(fTest, "%%%d.%dlf\0", field, mDigit);
	sprintf(mTest, fTest, value);
	max = (mTest[0] == (char) '-' || mTest[0] == ' ') ? (field) : (field - 1);
	if ((n = strlen(mTest)) > max) if ((mDigit -= n - max) < 0) mDigit = 0;
	return(mDigit);
}
