/******************************************************************************
	disl: displacement modeling by dislocations 

    ================================================================= 
   | (c) Copyright 1992 The Regents of the University of California. |
   |     All Rights Reserved.                                        |
   |                                                                 |
   |     This work was produced under the sponsorship of the         |
   |     United States Department of Energy.                         |
   |     The Government retains certain rights therein.              |
    ================================================================= 

	usage:   disl [optional parameters] < inputFile > displacementFile

	compile: cc -g -o disl disl.c getdisl.c readinput.a -lm -lget

	disl computes the vertical and horizontal deformation and line-length change 
	at longitude and latitude points (benchmarks) on the earth
	using the dislocation models of chinnery, BSSA 1961 (#3), 
	savage and hastie, JGR 1966 (#20), mansinha and smylie, JGR 1967 (#18) 
	and mansinha and smylie, BSSA 1971, (#5).
	a file containing the name of all parameters, faults, and benchmarks
	is used as input although separate fault and data files can be used.
	points can be entered as longitude/latitude/elevation or x/y/z.

	author: shawn larsen
	date:   may 1988
	where:  caltech
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "disl.h"
#include "usage.h"

struct data_group {
	faultType            *faults;
	benchMarkGroup           *bm;
	lineLengthGroup          *ll;
	profileGroup             *pf;
	contourGroup             *ct;
	char             **fileLines;
	};

typedef struct data_group Data;
typedef struct parameter_dislocation parameters;

main(argc, argv)
int    argc;
char **argv;
{
	Data        data[1];
	parameters   par[1];

	
	/* check for no input (print usage message) */
	if (argc == 1) if (checkUsage()) exit(0);


	/* get initial parameters */
	setpar(argc, argv);
	getParameter(par);
	endpar();


	/* read input file */
	readInput(par, data);


	/* get input parameters and data */
	getInput(par, data);


	/* form displacements */
	getModel(par, data);


	/* write output */
	writeOutput(par, data);


	/* free memory */
	freeMemory(par, data);
	
}






/******************************************************************************
	checkUsage: check usage and print usage message if no input

	usage: checkUsage()

	return 0: correct usage - continue execution
	       1: print   usage - stop     execution
******************************************************************************/
checkUsage()
{
	struct stat statbuf;
	int i;

	/* get input status */
	fstat(0, &statbuf);

	/* return if normal input */
	if (statbuf.st_mode & REGULAR_FILE) return(0);
	if (statbuf.st_mode & PIPE)         return(0);

	/* print usage and return exit */
	i = 0;
	while(dislUsage[i]) fprintf(stderr, "%s", dislUsage[i++]);
	return(1);

}




/******************************************************************************
   freeMemory: free memory

   usage: freeMemory(par, data)

   par  - parameters
   data - input data
******************************************************************************/
freeMemory(par, data)
parameters  *par;
Data       *data;
{
   int i;
	if (par->f) free(data->faults);
	for(i = 0; i < par->b; i++) free(data->bm[i].bm);
	for(i = 0; i < par->l; i++) free(data->ll[i].ll);
	for(i = 0; i < par->p; i++) free(data->pf[i].bm);
	for(i = 0; i < par->c; i++) free(data->ct[i].bm);
	if (par->b) free(data->bm);
	if (par->l) free(data->ll);
	if (par->p) free(data->pf);
	if (par->c) free(data->ct);
   for(i = 0; i < par->n; i++) free(data->fileLines[i])  ;free(data->fileLines);}




/******************************************************************************
	getParameter: read program parameters

	usage: getParameter(par)

	par - parameters
******************************************************************************/
getParameter(par)
parameters *par;
{
	int i;

	/* initialize parameters */
	par->XY           = XYFLAG;
	par->mDigit       = MDIGIT;
	par->residual     = RESFLAG;
	par->file[0]      = (char) '\0';
	par->dataFile[0]  = (char) '\0';
	par->faultFile[0] = (char) '\0';

	/* get parameters */
	getpar("xy"          , "d", &par->XY); 
	getpar("mdigit"      , "d", &par->mDigit); 
	getpar("residual"    , "d", &par->residual);
	getpar("file"        , "s",  par->file);
	getpar("datafile"    , "s",  par->dataFile);
	getpar("faultfile"   , "s",  par->faultFile);

	/* check parameters */
	if (par->mDigit <          0) par->mDigit = 0;
	if (par->mDigit >          9) par->mDigit = 9;

	/* check if parameter match from readline */
	if (getType("COMMENT") - LINECOMMENT) 
		fprintf(stderr, "WARNING: readline mismatch - LINECOMMENT\n");
	if (getType("NULL")    - LINENULL)    
		fprintf(stderr, "WARNING: readline mismatch - LINENULL\n");
	if (getType("BLANK") - LINEBLANK) 
		fprintf(stderr, "WARNING: readline mismatch - LINEBLANK\n");
	if (getType("VALUE") - LINEVALUE) 
		fprintf(stderr, "WARNING: readline mismatch - LINEVALUE\n");
	if (getType("PAR") - LINEPAR) 
		fprintf(stderr, "WARNING: readline mismatch - LINEPAR\n");
	if (getType("UNKNOWN") - LINEUNKNOWN) 
		fprintf(stderr, "WARNING: readline mismatch - LINEUNKNOWN\n");

}




/******************************************************************************
   readInput: read input files and store in argv array

   usage: readInput(par, data)

   par    - parameters
   data   - information
******************************************************************************/
readInput(par, data)
parameters *par;
Data       *data;
{
   int n, maxLines;
   unsigned size;

   /* initialize */
   maxLines = FILELINES;
   n = 0;

   /* read input file and store */
   size = maxLines*sizeof(char *);
   data->fileLines = (char **) malloc(size);
   while((n += argvFile(data->fileLines + n, FILELINES, stdin)) >= maxLines) {
      maxLines += FILELINES;
      size = (unsigned) maxLines*sizeof(char *);
      data->fileLines = (char **) realloc((char *) data->fileLines, size);
      };
   par->n = n;
      
   /* null last pointer */
   data->fileLines[par->n] = NULL;
      
}




/******************************************************************************
	getInput: get input parameters and data (allocate memory)

	usage: getInput(par, data)

	par    - parameters
	data   - information
******************************************************************************/
getInput(par, data)
parameters   *par;
Data        *data;
{
	lineLengthGroup   *ll;
	benchMarkGroup    *bm;
	profileGroup      *pf;
	contourGroup      *ct;
	int lineType, f;
	char **line, inLine[258];
	FILE *inFile, *getinFile();

	/* set line types */
	setLine("fault ",        LINEFAULT  );
	setLine("benchmarks ",   LINEBM     );
	setLine("linelengths ",  LINELENGTH );
	setLine("profile ",      LINEPROFILE);
	setLine("contour ",      LINECONTOUR);
	setLine("options ",      LINEOPTION );
	setLine("stop ",         LINESTOP   );
	setLine("BENCHMARKS ",   LINEBM     );
	setLine("LINELENGTHS ",  LINELENGTH );
	setLine("PROFILE ",      LINEBM     );
	setLine("CONTOUR ",      LINEBM     );

	/* count number of data items */
	par->b = 0; par->f = 0; par->l = 0; par->c = 0; par->p = 0;
	line = data->fileLines;
	while(*line != NULL) {
		switch((lineType = getLine(*line))) {
			case LINEFAULT:
				par->f++;
				break;
			case LINEBM:
				par->b++;
				break;
			case LINELENGTH:
				par->l++;
				break;
			case LINEPROFILE:
				par->p++;
				break;
			case LINECONTOUR:
				par->c++;
				break;
			default:
				break;
			}
		line++;
		}

	/* get memory */
	data->faults = (faultType       *) malloc(par->f*sizeof(faultType));
	data->bm     = (benchMarkGroup  *) malloc(par->b*sizeof(benchMarkGroup));
	data->ll     = (lineLengthGroup *) malloc(par->l*sizeof(lineLengthGroup));
	data->pf     = (profileGroup    *) malloc(par->p*sizeof(profileGroup));
	data->ct     = (contourGroup    *) malloc(par->c*sizeof(contourGroup));

	/* read in faults and data groups */
	f = 0;
	bm = NULL; ll = NULL; ct = NULL; pf = NULL;
	line = data->fileLines;
	while(*line != NULL) {
		switch(getLine(*line)) {
			case LINEFAULT:
				f + = getFault(data->faults + f, *line);
				lineType = LINEFAULT;
				break;
			case LINEBM:
				bm = (bm) ? (bm + 1) : (data->bm);
				bm->n = 0;
				if ((inFile = getinFile(*line))) {
					while(readLine(inLine, 258, inFile)) {
						if (getLine(inLine) == LINESTOP) break;
						if (getLine(inLine) == LINEDATA) {
							checkBMMemory(bm, FILELINES);
							bm->n += getBM(bm->bm + bm->n, inLine);
							}
						}
					fclose(inFile);
					}
				lineType = LINEBM;
				break;
			case LINELENGTH:
				ll = (ll) ? (ll + 1) : (data->ll);
				ll->n = 0;
				if ((inFile = getinFile(*line))) {
					while(readLine(inLine, 258, inFile)) {
						if (getLine(inLine) == LINESTOP) break;
						if (getLine(inLine) == LINEDATA) {
							checkLLMemory(ll, FILELINES);
							ll->n += getLL(ll->ll + ll->n, inLine);
							}
						}
					fclose(inFile);
					}
				lineType = LINELENGTH;
				break;
			case LINEDATA:
				if (lineType == LINEBM    ) {
					checkBMMemory(bm, FILELINES);
					bm->n += getBM(bm->bm + bm->n, *line);
					}
				else if (lineType == LINELENGTH) {
					checkLLMemory(ll, FILELINES);
					ll->n += getLL(ll->ll + ll->n, *line);
					}
				else if (bm == NULL && ll == NULL) {
					if (par->b++) free(data->bm);
					bm = (benchMarkGroup *) malloc(par->b*sizeof(benchMarkGroup));
					data->bm = bm;
					bm->n = 0;
					checkBMMemory(bm, FILELINES);
					bm->n += getBM(bm->bm + bm->n, *line);
					lineType = LINEBM;
					}
				break;
			case LINEPROFILE:
				pf = (pf) ? (pf + 1) : (data->pf);
				setArg(*line);
				mstArg("n", "d", &pf->n);
				endArg();
				if (pf->n >= 0) checkPFMemory(pf, pf->n);
				pf->n = getProfile(pf->bm, *line);
				strcpy(pf->line, *line);
				lineType = LINEPROFILE;
				break;
			case LINECONTOUR:
				ct = (ct) ? (ct + 1) : (data->ct);
				setArg(*line);
				mstArg("nx", "d", &ct->nx); 
				mstArg("ny", "d", &ct->ny); 
				endArg();
				if ((ct->n = ct->nx*ct->ny) >= 0) checkCTMemory(ct, ct->n);
				ct->n = getContour(ct->bm, *line);
				strcpy(ct->line, *line);
				lineType = LINECONTOUR;
				break;
			case LINEOPTION:
			case LINEPAR:
				setArg(*line);
				getOption(par, *line);
				endArg();
				lineType = LINEPAR;
				break;
			case LINESTOP:
				lineType = LINESTOP;
				break;
         case LINECOMMENT:
         case LINEBLANK:
         case LINENULL:
         case LINEUNKNOWN:
         default:
            break;
			}
		line++;
		}

	/* end line input */
	endLine();

}




FILE *
getinFile(line)
char *line;
{
	char fileName[82];
	FILE *inFile = NULL;
	fileName[0] = (char) '\0'; 
	setArg(line);
	if (getArg("file", "s", fileName)) {
		if (*fileName) {
			inFile = fopen(fileName, "r");
			if (inFile == NULL) {
				fprintf(stderr, "Error: unable to open file \"%s\"\n", fileName);
				exit(-1);
				}
			}
		}
	endArg(line);
	return(inFile);
}




checkBMMemory(bm, n)
benchMarkGroup *bm;
int n;
{
	unsigned size;
	if (bm->n == 0) {
		bm->bm = (benchMark *) malloc(n*sizeof(benchMark));
		}
	else if (bm->n%n == 0) {
		size = (unsigned) (bm->n + n)*sizeof(benchMark);
		bm->bm = (benchMark *) realloc((char *) bm->bm, size);
		}
}




checkLLMemory(ll, n)
lineLengthGroup *ll;
int n;
{
	unsigned size;
	if (ll->n == 0) {
		ll->ll = (lineLength *) malloc(n*sizeof(lineLength));
		}
	else if (ll->n%n == 0) {
		size = (unsigned) (ll->n + n)*sizeof(lineLength);
		ll->ll = (lineLength *) realloc((char *) ll->ll, size);
		}
}




checkPFMemory(pf, n)
profileGroup *pf;
int n;
{
	pf->bm = (benchMark *) malloc(n*sizeof(benchMark));
}




checkCTMemory(ct, n)
contourGroup *ct;
int n;
{
	ct->bm = (benchMark *) malloc(n*sizeof(benchMark));
}




/******************************************************************************
	getFault: get fault information from line

	usage: f = getFault(fault, line)

	fault - fault information
	line  - input line

	return: number of faults created (1)
******************************************************************************/
int
getFault(fault, line)
faultType *fault;
char      *line;
{
	int input;
	char lineType[20], fType[20];
	input = sscanf(line, "%s %s %lf %lf %lf %lf %lf %lf %lf %lf %s %d %d", 
	               lineType,      fType,
	               &fault->ds,   &fault->dl,   &fault->dt,   &fault->dw, 
	               &fault->dp,   &fault->xLon, &fault->yLat, &fault->az,
	                fault->name, &fault->nx,   &fault->ny);
	fault->ft = *fType;
	fault->name[7] = '\0';
	if (input == 9) {
		fault->nx = 1;
		fault->ny = 1;
		} 
	else if (input != 11) {
		fprintf(stderr, "ERROR: bad fault file format\n");
		fprintf(stderr, "\"%s\"\n", line);
		exit(-1);
		}
	if (fault->ft != (char) 's' && fault->ft != (char) 'd' &&
	    fault->ft != (char) 'S' && fault->ft != (char) 'D') {
		fprintf(stderr, "ERROR: %c is not a valid fault type\n", fault->ft);
		exit(-1);
		}
	if (fault->dp > 90. || fault->dp <= 0.) {
		fprintf(stderr, "ERROR: dip (%f) must be > 0. and <= 90.\n", fault->dp);
		exit(-1);
		}
	return(1);
}




/******************************************************************************
	getBM: get benchmark data from line (errors optional)

	usage: b = getBM(bm, inLine)

	bm       - benchmark information
	inLine   - input line

	return: number of benchmarks read (1)
******************************************************************************/
int
getBM(bm, inLine)
benchMark *bm;
char      *inLine;
{
	char line[256];
	strcpy(line, inLine);
	expandLine(line, 80);
	bm->xLon = 0.; bm->yLat = 0.; bm->zEle = 0.;
	bm->xDat = 0.; bm->yDat = 0.; bm->zDat = 0.;
	bm->xMod = 0.; bm->yMod = 0.; bm->zMod = 0.;
	strncpy(bm->name, line + 24, 7); bm->name[7] = (char) '\0';
	sscanf(line +  0, "%lf %lf %lf", &bm->xLon, &bm->yLat, &bm->zEle);
	sscanf(line + 31, "%lf %lf %lf", &bm->xDat, &bm->yDat, &bm->zDat);
	sscanf(line + 55, "%lf %lf %lf", &bm->xMod, &bm->yMod, &bm->zMod);
	if (line[17] != (char) ' ') sscanf(line + 17, "%lf", &bm->zEle);
	return(1);
}




/******************************************************************************
	getLL: get line-length data from line (errors optional)

	usage: l = getLL(ll, inLine)

	ll       - line-length information
	inLine   - input line

	return: number of line-lengths read (1)
******************************************************************************/
int
getLL(ll, inLine)
lineLength *ll;
char       *inLine;
{
	char line[256];
	strcpy(line, inLine);
	expandLine(line, 80);
	ll->xLon1 = 0.; ll->yLat1 = 0.; ll->zEle1 = 0.;
	ll->xLon2 = 0.; ll->yLat2 = 0.; ll->zEle2 = 0.;
	ll->dDat  = 0.; ll->dMod  = 0.;
	strncpy(ll->name, line + 48, 15); ll->name[15] = (char) '\0';
	sscanf(line +  0, "%lf %lf %lf", &ll->xLon1, &ll->yLat1, &ll->zEle1);
	sscanf(line + 24, "%lf %lf %lf", &ll->xLon2, &ll->yLat2, &ll->zEle2);
	sscanf(line + 63, "%lf", &ll->dDat);
	sscanf(line + 71, "%lf", &ll->dMod);
	if (line[17] != (char) ' ') sscanf(line + 17, "%lf", &ll->zEle1);
	if (line[41] != (char) ' ') sscanf(line + 41, "%lf", &ll->zEle2);
	return(1);
}




/******************************************************************************
	getProfile: get benchmark data along profile

	usage: n = getProfile(bm, line)

	bm   - benchmark information
	line - input line

	return: number of profile points created
******************************************************************************/
int
getProfile(bm, line)
benchMark  *bm;
char       *line;
{
	double xmin, xmax, ymin, ymax, dx, dy;
	int badInput, i, j, n, nxy;
	char name[8];

	/* initialize */
	badInput = 0;
	n = 0;

	/* get line arguments */
	setArg(line);
	if (!getArg("x1",   "F", &xmin)) badInput++;
	if (!getArg("x2",   "F", &xmax)) badInput++;
	if (!getArg("y1",   "F", &ymin)) badInput++;
	if (!getArg("y2",   "F", &ymax)) badInput++;
	if (!getArg("n",    "d", &nxy )) badInput++; 
	if (!getArg("name", "s",  name)) strncpy(name, "profile", 7);
	endArg();

	/* check if input is correct */
	if (nxy  <     2) badInput++;
	if (badInput) {
		fprintf(stderr, "ERROR: profile input line incorrect\n"); 
		exit(-1);
		}

	/* calculate limits */
	dx = (xmax - xmin)/(nxy - 1);
	dy = (ymax - ymin)/(nxy - 1);

	/* generate profile points from (xmin, ymin) to (xmax, ymax) */
	for(i = 0; i < nxy; i++) {
		strncpy(bm[n].name, name, 7); bm[n].name[7] = (char) '\0';
		bm[n].xDat = 0.;
		bm[n].yDat = 0.;
		bm[n].zDat = 0.;
		bm[n].xLon = xmin + (double) i*dx;
		bm[n].yLat = ymin + (double) i*dy;
		bm[n].zEle = 0.; 
		n++;
		}

	/* return number of profile points created */
	return(n);

}




/******************************************************************************
	getContour: get contour benchmark data 

	usage: n = getContour(bm, line)

	bm   - benchmark information
	line - input line

	return: number of contour points
******************************************************************************/
int
getContour(bm, line)
benchMark  *bm;
char       *line;
{
	double xmin, xmax, ymin, ymax, dx, dy;
	int badInput, i, j, n, nx, ny;
	char name[8];

	/* initialize */
	n = 0;
	badInput = 0;

	/* get line arguments */
	setArg(line);
	if (!getArg("x1",   "F", &xmin)) badInput++;
	if (!getArg("x2",   "F", &xmax)) badInput++;
	if (!getArg("y1",   "F", &ymin)) badInput++;
	if (!getArg("y2",   "F", &ymax)) badInput++;
	if (!getArg("nx",   "d", &nx  )) badInput++; 
	if (!getArg("ny",   "d", &ny  )) badInput++; 
	if (!getArg("name", "s",  name)) strncpy(name, "contour", 7);
	endArg();

	/* check if input is correct */
	if (nx   <     2) badInput++;
	if (ny   <     2) badInput++;
	if (badInput) {
		fprintf(stderr, "ERROR: contour input line incorrect\n"); 
		exit(-1);
		}

	/* calculate limits */
	dx = (xmax - xmin)/(nx - 1);
	dy = (ymax - ymin)/(ny - 1);

	/* generate contour points row-by-row from (xmin, ymin) */
	for(j = 0; j < ny; j++) 
		for(i = 0; i < nx; i++) {
			strncpy(bm[n].name, name, 7); bm[n].name[7] = (char) '\0';
			bm[n].xDat = 0.;
			bm[n].yDat = 0.;
			bm[n].zDat = 0.;
			bm[n].xLon = xmin + (double) i*dx;
			bm[n].yLat = ymin + (double) j*dy;
			bm[n].zEle = 0.; 
			n++;
			}

	/* return number of contour points created */
	return(n);

}




/******************************************************************************
	getOption: get parameter information from line

	usage: getOption(par, line)

	par  - parameters
	line - input line containing options or parameters
******************************************************************************/
int
getOption(par, line)
parameters *par;
char       *line;
{
	setArg(line);
	getArg("xy"    ,   "d", &par->XY); 
	getArg("mdigit",   "d", &par->mDigit); 
	getArg("residual", "d", &par->residual); 
	endArg();
}




/******************************************************************************
	getModel: get displacements

	usage: getModel(par, data)

	par    - parameters
	data   - input data
******************************************************************************/
getModel(par, data)
parameters *par;
Data       *data;
{
	int i;
	for(i = 0; i < par->b; i++) 
		getModelBM(data->faults, data->bm[i].bm, par->XY, data->bm[i].n, par->f);
	for(i = 0; i < par->p; i++) 
		getModelBM(data->faults, data->pf[i].bm, par->XY, data->pf[i].n, par->f);
	for(i = 0; i < par->c; i++) 
		getModelBM(data->faults, data->ct[i].bm, par->XY, data->ct[i].n, par->f);
	for(i = 0; i < par->l; i++)
		getModelLL(data->faults, data->ll[i].ll, par->XY, data->ll[i].n, par->f);
}



getModelBM(faults, bm, XY, n, f)
faultType *faults;
benchMark     *bm;
int f, n, XY;
{
	double x, y, Ux, Uy, Uz;
	int i, j;

	/* loop over benchmarks and faults */
	for(i = 0; i < n; i++) {
		bm[i].xMod = 0.; 
		bm[i].yMod = 0.; 
		bm[i].zMod = 0.; 
		for(j = 0; j < f; j++) {
			getFaultCoord(&x, &y, bm + i, faults + j, XY);
			getDisl(&Ux, &Uy, &Uz, x, y, faults + j); 
			bm[i].xMod += Ux*faults[j].ds;
			bm[i].yMod += Uy*faults[j].ds;
			bm[i].zMod += Uz*faults[j].ds;
			}
		}

}



getModelLL(faults, ll, XY, n, f)
faultType *faults;
lineLength    *ll;
int f, n, XY;
{
	benchMark bm[2];
	double x, y, Ux, Uy, Uz, dx, dy, dz, lengthOld, lengthNew;
	int i, j;

	/* loop over line-length stations and faults */
	for(i = 0; i < n; i++) {
		bm[0].xLon = ll[i].xLon1; bm[0].xMod = 0.; 
		bm[0].yLat = ll[i].yLat1; bm[0].yMod = 0.; 
		bm[0].zEle = ll[i].zEle1; bm[0].zMod = 0.; 
		bm[1].xLon = ll[i].xLon2; bm[1].xMod = 0.; 
		bm[1].yLat = ll[i].yLat2; bm[1].yMod = 0.; 
		bm[1].zEle = ll[i].zEle2; bm[1].zMod = 0.; 
		dx = bm[1].xLon - bm[0].xLon;
		dy = bm[1].yLat - bm[0].yLat;
		dz = bm[0].zEle - bm[1].zEle;
		if (!XY) {
			dx =  dx*(KMD)*cos(((bm[0].yLat + bm[1].yLat)/2.)/(RAD));
			dy =  dy*(KMD);
			dz = -dz/1000.;
			}
		dx *= 100000.; dy *= 100000.; dz *= 100000.; 
		lengthOld = sqrt(dx*dx + dy*dy + dz*dz);
		for(j = 0; j < f; j++) {
			getFaultCoord(&x, &y, bm + 0, faults + j, XY);
			getDisl(&Ux, &Uy, &Uz, x, y, faults + j); 
			bm[0].xMod += Ux*faults[j].ds;
			bm[0].yMod += Uy*faults[j].ds;
			bm[0].zMod += Uz*faults[j].ds;
			getFaultCoord(&x, &y, bm + 1, faults + j, XY);
			getDisl(&Ux, &Uy, &Uz, x, y, faults + j); 
			bm[1].xMod += Ux*faults[j].ds;
			bm[1].yMod += Uy*faults[j].ds;
			bm[1].zMod += Uz*faults[j].ds;
			}
		dx += bm[1].xMod - bm[0].xMod;
		dy += bm[1].yMod - bm[0].yMod;
		dz += bm[1].zMod - bm[0].zMod;
		lengthNew = sqrt(dx*dx + dy*dy + dz*dz);
		ll[i].dMod = lengthNew - lengthOld;
		}

}



getFaultCoord(x, y, bm, fault, XY)
benchMark  *bm;
faultType  *fault;
double *x, *y;
int XY;
{
	double cs, sn, dx, dy;
	cs = cos(fault->az/(RAD));
	sn = sin(fault->az/(RAD));
	dx = bm->xLon - fault->xLon;
	dy = bm->yLat - fault->yLat;
	if (!XY) {
		dx = dx*(KMD)*cos(fault->yLat/(RAD));
		dy = dy*(KMD);
		}
	*x = cs*dx + sn*dy;
	*y = cs*dy - sn*dx;
	if (fault->ft == (char) 'S' || fault->ft == (char) 'D') {
		if (fault->dp <= 89.) *y += fault->dt/tan(fault->dp/(RAD));
		}
}



getDisl(ux, uy, uz, x, y, f)
faultType *f;
double *ux, *uy, *uz, x, y;
{
	double fx, fy, z, cs, sn;

	/* get displacements at surface (depth = 0) */
	z = 0.;

	/* compute depth to top of fault */
/*
	dt = f->dt;
	if (f->ft == (char) 'S' || f->ft == (char) 'D') {
		dt -= (f->dw/2.)*sin(fabs(f->dp)/(RAD));
		}
*/

	/* get displacements in fault coordinate system */
	getdisl(x, y, z, f->dl, f->dt, f->dw, f->dp, f->ft, &fx, &fy, uz);

	/* rotate displacements from fault to Cartesian coordinates */
	cs = cos(f->az/(RAD));
	sn = sin(f->az/(RAD));
	*ux = fx*cs - fy*sn;
	*uy = fx*sn + fy*cs;
	
}



/******************************************************************************
	writeOutput: write results to standard output

	usage: writeOutput(par, data);

	par    - parameters
	data   - input data
******************************************************************************/
writeOutput(par, data)
parameters  *par;
Data       *data;
{
	benchMark  *bm;
	lineLength *ll;
	faultType  *f;
	int i, n;
	char dFormat[82], lFormat[82], dTitle[200];

	/* initialize */
	f = data->faults;

	/* write header and input parameters */
	printf("                               DISLOCATION MODELING\n");
	printf("                               --------------------\n");
	printf("options xy=%-d mdigit=%-d residual=%-d\n", 
	        par->XY, par->mDigit, par->residual);


	/* write header information for faults */
   printf("\n");
   printf("\n");
	printf("                                  FAULTS MODELED\n");
	printf("                                  --------------\n");
	printf("\n");
	if (par->XY) {
		printf("    Ftype   Slip  Length  Depth  Width   Dip ");
		printf("    XDist    YDist Azimuth Name\n");
		printf("    -----   ----  ------  -----  -----   --- ");
		printf("    -----    ----- ------- ----\n");
		}
	else {
		printf("    Ftype  Slip   Length  Depth  Width   Dip ");
		printf("Longitude Latitude Azimuth Name\n");
		printf("    -----  ----   ------  -----  -----   --- ");
		printf("--------- -------- ------- ----\n");
		}

   /* write faults */
   for(i = 0; i < par->f; i++) {
      printf("fault ");
      printf("%c %8.2lf %7.2lf %6.2lf %6.2lf %5.1lf %9.4lf %8.4lf %6.1lf  ",
              f[i].ft, f[i].ds, f[i].dl, f[i].dt, f[i].dw, f[i].dp,
              f[i].xLon, f[i].yLat, f[i].az);
      printf("%-7s",  f[i].name);
      printf("\n");
      }

	/* write header information for benchmarks */
   printf("\n");
   printf("\n");
   printf("                                    DISPLACEMENT\n");
   printf("                                    ------------\n");

	/* get default format strings */
	sprintf(dFormat, "%%9. lf%%8. lf%%7. lf%%+7s%%8. lf%%8. lf%%8. lf");
	sprintf(dFormat + 40, "%%8. lf%%8. lf%%8. lf\n\0");
	sprintf(lFormat, "%%9. lf%%8. lf%%7. lf%%9. lf%%8. lf%%7. lf");
	sprintf(lFormat + 36, "%%15s%%8. lf%%8. lf\n\0");

	/* write benchmark information to output file */
	for(i = 0; i < par->b; i++) {
		bm = data->bm[i].bm;
		printDLegend(par->XY, par->residual, NULL, LINEBM); 
		if (par->residual) for(n = 0; n < data->bm[i].n; n++) getBMResidual(bm+n);
		for(n = 0; n < data->bm[i].n; n++) {
			getDFormat(dFormat, par, bm + n);
			printf(dFormat, bm[n].xLon, bm[n].yLat, bm[n].zEle, bm[n].name,
		               	 bm[n].xDat, bm[n].yDat, bm[n].zDat, 
		               	 bm[n].xMod, bm[n].yMod, bm[n].zMod);
			};
		}

	/* write line-length information to output file */
	for(i = 0; i < par->l; i++) {
		ll = data->ll[i].ll;
		printDLegend(par->XY, par->residual, NULL, LINELENGTH); 
		if (par->residual) for(n = 0; n < data->bm[i].n; n++) getLLResidual(ll+n);
		for(n = 0; n < data->ll[i].n; n++) {
			getLFormat(lFormat, par, ll + n);
			printf(lFormat, ll[n].xLon1, ll[n].yLat1, ll[n].zEle1,
			                ll[n].xLon2, ll[n].yLat2, ll[n].zEle2, 
		               	 ll[n].name , ll[n].dDat,  ll[n].dMod);
			}
		}

	/* write profile information to output file */
	for(i = 0; i < par->p; i++) {
		bm = data->pf[i].bm;
		printDLegend(par->XY, par->residual, data->pf[i].line, LINEPROFILE); 
		if (par->residual) for(n = 0; n < data->bm[i].n; n++) getBMResidual(bm+n);
		for(n = 0; n < data->pf[i].n; n++) {
			getDFormat(dFormat, par, bm + n);
			printf(dFormat, bm[n].xLon, bm[n].yLat, bm[n].zEle, bm[n].name,
		               	 bm[n].xDat, bm[n].yDat, bm[n].zDat, 
		               	 bm[n].xMod, bm[n].yMod, bm[n].zMod);
			}
		}

	/* write contour information to output file */
	for(i = 0; i < par->c; i++) {
		bm = data->ct[i].bm;
		printDLegend(par->XY, par->residual, data->ct[i].line, LINECONTOUR); 
		if (par->residual) for(n = 0; n < data->bm[i].n; n++) getBMResidual(bm+n);
		for(n = 0; n < data->ct[i].n; n++) {
			getDFormat(dFormat, par, bm + n);
			printf(dFormat, bm[n].xLon, bm[n].yLat, bm[n].zEle, bm[n].name,
		               	 bm[n].xDat, bm[n].yDat, bm[n].zDat, 
		               	 bm[n].xMod, bm[n].yMod, bm[n].zMod);
			}
		}

}




getBMResidual(bm)
benchMark *bm;
{
	bm->xMod = bm->xDat - bm->xMod;
	bm->yMod = bm->yDat - bm->yMod;
	bm->zMod = bm->zDat - bm->zMod;
}




getLLResidual(ll)
lineLength *ll;
{
	ll->dMod = ll->dDat - ll->dMod;
}




printDLegend(XY, residual, line, mode)
int XY, residual, mode;
char *line;
{
	int n = 0; 
	char string[258];
	switch(mode) {
		case LINEBM:
			printf("\n");
			printf("BENCHMARKS\n");
			printf("\n");
			printDLabel(XY, residual);
			break;
		case LINELENGTH:
			printf("\n");
			printf("LINELENGTHS\n");
			printf("\n");
			printLLabel(XY, residual);
			break;
		case LINEPROFILE:
			while(line[n] == (char) ' ') n++;
			printf("\n\n");
			strcpy(string, line + n);
			strncpy(string, "PROFILE", 7);
			printf("%s\n", string);
			printf("\n");
			printDLabel(XY, residual);
			break;
		case LINECONTOUR:
			while(line[n] == (char) ' ') n++;
			printf("\n\n");
			strcpy(string, line + n);
			strncpy(string, "CONTOUR", 7);
			printf("%s\n", string);
			printf("\n");
			printDLabel(XY, residual);
			break;
		}
}




printDLabel(XY, residual)
int XY, residual;
{
   if (XY) {
		printf("    XDIST   YDIST  DEPTH   NAME    XDAT    YDAT    ZDAT    ");
		if (residual) 
		printf("XRES    YRES    ZRES\n");
		else
		printf("XMOD    YMOD    ZMOD\n");
		printf("    -----   -----  -----   ----    ----    ----    ----    ");
		printf("----    ----    ----\n");
      }
   else {
		printf("LONGITUDE LATITUDE  ELEV   NAME    EDAT    NDAT    VDAT    ");
		if (residual)
		printf("ERES    NRES    VRES\n");
		else
		printf("EMOD    NMOD    VMOD\n");
		printf("--------- --------  ----   ----    ----    ----    ----    ");
		printf("----    ----    ----\n");
      }
}




printLLabel(XY, residual)
int XY, residual;
{
   if (XY) {
		printf("    XDIST   YDIST  DEPTH    XDIST   YDIST  DEPTH     ");
		if (residual)
		printf("NAME          DDAT    DRES\n");
		else
		printf("NAME          DDAT    DMOD\n");
		printf("    -----   -----  -----    -----   -----  -----     ");
		printf("----          ----    ----\n");
      }
   else {
		printf("LONGITUDE LATITUDE ELEV LONGITUDE LATITUDE ELEV      ");
		if (residual)
		printf("NAME          DDAT    DRES\n");
		else
		printf("NAME          DDAT    DMOD\n");
		printf("--------- -------- ---- --------- -------- ----      ");
		printf("----          ----    ----\n");
      }
}




getDFormat(format, par, bm)
parameters *par;
benchMark  *bm;
char       *format;
{
   format[ 3] = (char) (getMDigit(bm->xLon, 9,           4) + '0');
   format[ 9] = (char) (getMDigit(bm->yLat, 8,           4) + '0');
   format[15] = (char) (getMDigit(bm->zEle, 7,           2) + '0');
   format[25] = (char) (getMDigit(bm->xDat, 8,           2) + '0');
   format[31] = (char) (getMDigit(bm->yDat, 8,           2) + '0');
   format[37] = (char) (getMDigit(bm->zDat, 8,           2) + '0');
   format[43] = (char) (getMDigit(bm->xMod, 8, par->mDigit) + '0');
   format[49] = (char) (getMDigit(bm->yMod, 8, par->mDigit) + '0');
   format[55] = (char) (getMDigit(bm->zMod, 8, par->mDigit) + '0');	
}




getLFormat(format, par, ll)
parameters *par;
lineLength *ll;
char       *format;
{
   format[ 3] = (char) (getMDigit(ll->xLon1, 9,           4) + '0');
   format[ 9] = (char) (getMDigit(ll->yLat1, 8,           4) + '0');
   format[15] = (char) (getMDigit(ll->zEle1, 7,           2) + '0');
   format[21] = (char) (getMDigit(ll->xLon2, 9,           4) + '0');
   format[27] = (char) (getMDigit(ll->yLat2, 8,           4) + '0');
   format[33] = (char) (getMDigit(ll->zEle2, 7,           2) + '0');
   format[43] = (char) (getMDigit(ll->dDat,  8,           2) + '0');
   format[49] = (char) (getMDigit(ll->dMod,  8, par->mDigit) + '0');
}




int
getMDigit(value, field, mDigit)
double value;
int    field, mDigit;
{
   int max, n;
   char fTest[20], mTest[20];
   sprintf(fTest, "%%%d.%dlf\0", field, mDigit);
   sprintf(mTest, fTest, value);
   max = (mTest[0] == (char) '-' || mTest[0] == ' ') ? (field) : (field - 1);
   if ((n = strlen(mTest)) > max) if ((mDigit -= n - max) < 0) mDigit = 0;
   return(mDigit);
}




expandLine(line, n)
int      n;
char *line;
{
	int i, j;
	for(i = 0; i < n; i++) {
		if (line[i] == (char) '\n' || line[i] == (char) '\0') {
			for(j = i; j < n; j++) line[j] = (char) ' ';
			line[n] = (char) '\0';
			}
		}
}
