/******************************************************************************
	dislflt: convert disl output into plotting format (mplot)

    ================================================================= 
   | (c) Copyright 1992 The Regents of the University of California. |
   |     All Rights Reserved.                                        |
   |                                                                 |
   |     This work was produced under the sponsorship of the         |
   |     United States Department of Energy.                         |
   |     The Government retains certain rights therein.              |
    ================================================================= 

	usage:   dislflt [optional parameters] < dislOutputFile > plotOutputFile

	compile: cc -g -o dislflt dislflt.c readinput.a -lm -lget

	dislflt coverts disl output files into a format that is used for plotting. 
	the plotting options are ...
	1) plot the station positions and/or line-lengths
	2) convert horizontal displacement at benchmarks to arrows
	in all cases output can be raw (e.g., arrows only) 
	or include mplot parameters.
	a separate mplot parameter file can be read to facilitate plotting.

	author: shawn larsen
	date:   december 1992
	where:  llnl
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "disl.h"
#include "usage.h"

struct data_group {
	faultType            *faults;
	benchMarkGroup           *bm;
	lineLengthGroup          *ll;
	profileGroup             *pf;
	contourGroup             *ct;
	char             **fileLines;
	};

typedef struct data_group Data;
typedef struct parameter_dislocation parameters;

main(argc, argv)
int    argc;
char **argv;
{
	Data        data[1];
	parameters   par[1];

	
	/* check for no input (print usage message) */
	if (argc == 1) if (checkUsage()) exit(0);


	/* get initial parameters */
	setpar(argc, argv);
	getParameter(par);
	endpar();


	/* read input file */
	readInput(par, data);


	/* get input parameters and data */
	getInput(par, data);


	/* write output */
	writeOutput(par, data);


	/* free memory */
	freeMemory(par, data);
	
}






/******************************************************************************
	checkUsage: check usage and print usage message if no input

	usage: checkUsage()

	return 0: correct usage - continue execution
	       1: print   usage - stop     execution
******************************************************************************/
checkUsage()
{
	struct stat statbuf;
	int i;

	/* get input status */
	fstat(0, &statbuf);

	/* return if normal input */
	if (statbuf.st_mode & REGULAR_FILE) return(0);
	if (statbuf.st_mode & PIPE)         return(0);

	/* print usage and return exit */
	i = 0;
	while(dislfltUsage[i]) fprintf(stderr, "%s", dislfltUsage[i++]);
	return(1);

}




/******************************************************************************
   freeMemory: free memory

   usage: freeMemory(par, data)

   par  - parameters
   data - input data
******************************************************************************/
freeMemory(par, data)
parameters  *par;
Data       *data;
{
   int i;
	if (par->f) free(data->faults);
	for(i = 0; i < par->b; i++) free(data->bm[i].bm);
	for(i = 0; i < par->l; i++) free(data->ll[i].ll);
	for(i = 0; i < par->p; i++) free(data->pf[i].bm);
	for(i = 0; i < par->c; i++) free(data->ct[i].bm);
	if (par->b) free(data->bm);
	if (par->l) free(data->ll);
	if (par->p) free(data->pf);
	if (par->c) free(data->ct);
   for(i = 0; i < par->n; i++) free(data->fileLines[i])  ;free(data->fileLines);}




/******************************************************************************
	getParameter: read program parameters

	usage: getParameter(par)

	par - parameters
******************************************************************************/
getParameter(par)
parameters *par;
{
	int i;

	/* initialize parameters */
	par->XY           = XYFLAG;
	par->residual     = RESFLAG;
	par->file[0]      = (char) '\0';

	/* get parameters */
	getpar("xy"          , "d", &par->XY); 
	getpar("residual"    , "d", &par->residual);
	getpar("file"        , "s",  par->file);

	/* check if parameter match from readline */
	if (getType("COMMENT") - LINECOMMENT) 
		fprintf(stderr, "WARNING: readline mismatch - LINECOMMENT\n");
	if (getType("NULL")    - LINENULL)    
		fprintf(stderr, "WARNING: readline mismatch - LINENULL\n");
	if (getType("BLANK") - LINEBLANK) 
		fprintf(stderr, "WARNING: readline mismatch - LINEBLANK\n");
	if (getType("VALUE") - LINEVALUE) 
		fprintf(stderr, "WARNING: readline mismatch - LINEVALUE\n");
	if (getType("PAR") - LINEPAR) 
		fprintf(stderr, "WARNING: readline mismatch - LINEPAR\n");
	if (getType("UNKNOWN") - LINEUNKNOWN) 
		fprintf(stderr, "WARNING: readline mismatch - LINEUNKNOWN\n");

}




/******************************************************************************
   readInput: read input files and store in argv array

   usage: readInput(par, data)

   par    - parameters
   data   - information
******************************************************************************/
readInput(par, data)
parameters *par;
Data       *data;
{
   int n, maxLines;
   unsigned size;

   /* initialize */
   maxLines = FILELINES;
   n = 0;

   /* read input file and store */
   size = maxLines*sizeof(char *);
   data->fileLines = (char **) malloc(size);
   while((n += argvFile(data->fileLines + n, FILELINES, stdin)) >= maxLines) {
      maxLines += FILELINES;
      size = (unsigned) maxLines*sizeof(char *);
      data->fileLines = (char **) realloc((char *) data->fileLines, size);
      };
   par->n = n;
      
   /* null last pointer */
   data->fileLines[par->n] = NULL;
      
}




/******************************************************************************
	getInput: get input parameters and data (allocate memory)

	usage: getInput(par, data)

	par    - parameters
	data   - information
******************************************************************************/
getInput(par, data)
parameters   *par;
Data        *data;
{
	lineLengthGroup   *ll;
	benchMarkGroup    *bm;
	profileGroup      *pf;
	contourGroup      *ct;
	int lineType, f;
	char **line, inLine[258];
	FILE *inFile, *getinFile();

	/* set line types */
	setLine("fault ",        LINEFAULT  );
	setLine("benchmarks ",   LINEBM     );
	setLine("linelengths ",  LINELENGTH );
	setLine("options ",      LINEOPTION );
	setLine("stop ",         LINESTOP   );
	setLine("BENCHMARKS ",   LINEBM     );
	setLine("LINELENGTHS ",  LINELENGTH );
	setLine("PROFILE ",      LINEBM     );
	setLine("CONTOUR ",      LINEBM     );

	/* count number of data items */
	par->b = 0; par->f = 0; par->l = 0; 
	line = data->fileLines;
	while(*line != NULL) {
		switch((lineType = getLine(*line))) {
			case LINEFAULT:
				par->f++;
				break;
			case LINEBM:
				par->b++;
				break;
			case LINELENGTH:
				par->l++;
				break;
			default:
				break;
			}
		line++;
		}

	/* get memory */
	data->faults = (faultType       *) malloc(par->f*sizeof(faultType));
	data->bm     = (benchMarkGroup  *) malloc(par->b*sizeof(benchMarkGroup));
	data->ll     = (lineLengthGroup *) malloc(par->l*sizeof(lineLengthGroup));

	/* read in faults and data groups */
	f = 0;
	bm = NULL; ll = NULL;
	line = data->fileLines;
	while(*line != NULL) {
		switch(getLine(*line)) {
			case LINEFAULT:
				f + = getFault(data->faults + f, *line);
				lineType = LINEFAULT;
				break;
			case LINEBM:
				bm = (bm) ? (bm + 1) : (data->bm);
				bm->n = 0;
				lineType = LINEBM;
				break;
			case LINELENGTH:
				ll = (ll) ? (ll + 1) : (data->ll);
				ll->n = 0;
				lineType = LINELENGTH;
				break;
			case LINEDATA:
				if (lineType == LINEBM    ) {
					checkBMMemory(bm, FILELINES);
					bm->n += getBM(bm->bm + bm->n, *line);
					}
				else if (lineType == LINELENGTH) {
					checkLLMemory(ll, FILELINES);
					ll->n += getLL(ll->ll + ll->n, *line);
					}
				else if (bm == NULL && ll == NULL) {
					if (par->b++) free(data->bm);
					bm = (benchMarkGroup *) malloc(par->b*sizeof(benchMarkGroup));
					data->bm = bm;
					bm->n = 0;
					checkBMMemory(bm, FILELINES);
					bm->n += getBM(bm->bm + bm->n, *line);
					lineType = LINEBM;
					}
				break;
			case LINEOPTION:
			case LINEPAR:
				setArg(*line);
				getOption(par, *line);
				endArg();
				lineType = LINEPAR;
				break;
			case LINESTOP:
				lineType = LINESTOP;
				break;
         case LINECOMMENT:
         case LINEBLANK:
         case LINENULL:
         case LINEUNKNOWN:
         default:
            break;
			}
		line++;
		}

	/* end line input */
	endLine();

}




checkBMMemory(bm, n)
benchMarkGroup *bm;
int n;
{
	unsigned size;
	if (bm->n == 0) {
		bm->bm = (benchMark *) malloc(n*sizeof(benchMark));
		}
	else if (bm->n%n == 0) {
		size = (unsigned) (bm->n + n)*sizeof(benchMark);
		bm->bm = (benchMark *) realloc((char *) bm->bm, size);
		}
}




checkLLMemory(ll, n)
lineLengthGroup *ll;
int n;
{
	unsigned size;
	if (ll->n == 0) {
		ll->ll = (lineLength *) malloc(n*sizeof(lineLength));
		}
	else if (ll->n%n == 0) {
		size = (unsigned) (ll->n + n)*sizeof(lineLength);
		ll->ll = (lineLength *) realloc((char *) ll->ll, size);
		}
}




/******************************************************************************
	getFault: get fault information from line

	usage: f = getFault(fault, line)

	fault - fault information
	line  - input line

	return: number of faults created (1)
******************************************************************************/
int
getFault(fault, line)
faultType *fault;
char      *line;
{
	int input;
	char lineType[20], fType[20];
	input = sscanf(line, "%s %s %lf %lf %lf %lf %lf %lf %lf %lf %s %d %d", 
	               lineType,      fType,
	               &fault->ds,   &fault->dl,   &fault->dt,   &fault->dw, 
	               &fault->dp,   &fault->xLon, &fault->yLat, &fault->az,
	                fault->name, &fault->nx,   &fault->ny);
	fault->ft = *fType;
	fault->name[7] = '\0';
	if (input == 9) {
		fault->nx = 1;
		fault->ny = 1;
		} 
	else if (input != 11) {
		fprintf(stderr, "ERROR: bad fault file format\n");
		fprintf(stderr, "\"%s\"\n", line);
		exit(-1);
		}
	return(1);
}




/******************************************************************************
	getBM: get benchmark data from line 

	usage: b = getBM(bm, line)

	bm       - benchmark information
	line     - input line

	return: number of benchmarks read (1)
******************************************************************************/
int
getBM(bm, line)
benchMark *bm;
char      *line;
{
	strncpy(bm->name, line + 24, 7); bm->name[7] = (char) '\0';
	sscanf(line +  0, "%lf %lf %lf", &bm->xLon, &bm->yLat, &bm->zEle);
	sscanf(line + 31, "%lf %lf %lf", &bm->xDat, &bm->yDat, &bm->zDat);
	sscanf(line + 55, "%lf %lf %lf", &bm->xMod, &bm->yMod, &bm->zMod);
	if (line[17] != (char) ' ') sscanf(line + 17, "%lf", &bm->zEle);
	return(1);
}




/******************************************************************************
	getLL: get line-length data from line

	usage: l = getLL(ll, line)

	ll       - line-length information
	line     - input line

	return: number of line-lengths read (1)
******************************************************************************/
int
getLL(ll, line)
lineLength *ll;
char       *line;
{
	strncpy(ll->name, line + 48, 15); ll->name[15] = (char) '\0';
	sscanf(line +  0, "%lf %lf %lf", &ll->xLon1, &ll->yLat1, &ll->zEle1);
	sscanf(line + 24, "%lf %lf %lf", &ll->xLon2, &ll->yLat2, &ll->zEle2);
	sscanf(line + 63, "%lf", &ll->dDat);
	sscanf(line + 71, "%lf", &ll->dMod);
	if (line[17] != (char) ' ') sscanf(line + 17, "%lf", &ll->zEle1);
	if (line[41] != (char) ' ') sscanf(line + 41, "%lf", &ll->zEle2);
	return(1);
}




/******************************************************************************
	getOption: get parameter information from line

	usage: getOption(par, line)

	par  - parameters
	line - input line containing options or parameters
******************************************************************************/
int
getOption(par, line)
parameters *par;
char       *line;
{
	setArg(line);
	getArg("xy"    ,   "d", &par->XY); 
	getArg("residual", "d", &par->residual); 
	endArg();
}




/******************************************************************************
	writeOutput: write results to standard output

	usage: writeOutput(par, data);

	par    - parameters
	data   - input data
******************************************************************************/
writeOutput(par, data)
parameters  *par;
Data       *data;
{
	benchMark  *bm;
	lineLength *ll;
	faultType  *f;
	int i, n;

	/* initialize */
	f = data->faults;

	/* determine minimum and maximum values */
	getLimits();

}




getLimits()
{
}




writeArrow()
{
}
