/******************************************************************************
	disn: strain/stress modeling by dislocations

	*******************************************************************
	* (c) Copyright 1992 The Regents of the University of California. *
	*     All Rights Reserved.                                        *
	*                                                                 *
	*     This work was produced under the sponsorship of the         *
	*     United States Department of Energy.                         *
	*     The Government retains certain rights therein.              *
	*******************************************************************

	usage:   disn [optional parameters] < inputFile > strainFile

	compile: cc -g -o disn disn.c getdisl.c readinput.a -lm -lget

	disn calculates the 3-dimensional strain/stress tensor at specified points
	using the dislocation models of chinnery, BSSA 1961 (#3), 
	savage and hastie, JGR 1966 (#20), mansinha and smylie, JGR 1967 (#18), 
	and mansinha and smylie, BSSA 1971 (#5).
	a file containing the name of all paramaters, faults, and 
	locations is used as input.
	points can be entered as longitude/latitude/depth or x/y/z.

	author: shawn larsen
	date:   july 1992
	where:  llnl
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "disl.h"
#include "usage.h"

struct data_group {
	strainTensor    *strain;
	stressTensor    *stress;
	faultType       *faults;
	PointGroup      *groups;
	char        **fileLines;
	};

typedef struct data_group Data;
typedef struct parameter_strainstress parameters;

main(argc, argv)
int    argc;
char **argv;
{
	Data          data[1];
	parameters     par[1];
	int i;

	
	/* check for no input (print usage message) */
	if (argc == 1) if (checkUsage()) exit(0);

	
	/* get initial parameters */
	setpar(argc, argv);
	getPar(par);
	endpar();


	/* read input file */
	readInput(par, data);


	/* get input parameters and data */
	getInput(par, data);


	/* form strain tensor */
	getStrain(par, data->strain, data->faults);


	/* convert strain to micro strain */
	if (!par->stress) getMicroStrain(par, data->strain);


	/* convert strain to stress (stress and strain point to same) */
	if (par->stress) getStress(par, data->stress);


	/* get stress components at specified orientation on plane or datum */
	if (par->stress == FAULTSTRESS) getFaultStress(par, data);


	/* write output (stress and strain point to same) */
	writeOutput(par, data);


	/* free memory */
	freeMemory(par, data);

}






/******************************************************************************
	checkUsage: check usage and print usage message if no input

	usage: checkUsage()

	return 0: correct usage - continue execution
	       1: print   usage - stop     execution
******************************************************************************/
checkUsage()
{
	struct stat statbuf;
	int i;
	double increment;
	int bmax, fmax, m, xy, strain;

	/* get input status */
	fstat(0, &statbuf);

	/* return if normal input */
	if (statbuf.st_mode & REGULAR_FILE) return(0);
	if (statbuf.st_mode & PIPE)         return(0);

	/* print usage and return exit */
	i = 0;
	while(disnUsage[i]) fprintf(stderr, "%s", disnUsage[i++]);
	return(1);

}




/******************************************************************************
	freeMemory: free memory

	usage: freeMemory(par, data)

	par  - parameters
	data - input data
******************************************************************************/
freeMemory(par, data)
parameters  *par;
Data       *data;
{
	int i;
	if (par->f) free(data->faults);
	if (par->b) free(data->strain);
	for(i = 0; i < par->g; i++) free(data->groups[i].line);free(data->groups);
	for(i = 0; i < par->n; i++) free(data->fileLines[i])  ;free(data->fileLines);
}




/******************************************************************************
	getPar: read program parameters

	usage: getPar(par)

	par - parameters
******************************************************************************/
getPar(par)
parameters *par;
{
	int i;

	/* initialize parameters */
	par->b            = 0;
	par->f            = 0;
	par->XY           = XYFLAG;
	par->stress       = STRESSFLAG;
	par->mDigit       = MDIGIT;
	par->increment    = INCREMENT;
	par->friction     = FRICTION;
	par->lamda        = LAMDA;
	par->mu           = MU;
	par->faultFile[0] = (char) '\0';

	/* get parameters */
	getpar("xy"          , "d", &par->XY); 
	getpar("stress"      , "d", &par->stress);
	getpar("mdigit"      , "d", &par->mDigit);
	getpar("increment"   , "F", &par->increment);
	getpar("faultfile"   , "s",  par->faultFile);

	/* check parameters */
	if (par->increment <         0.) par->increment = INCREMENT;
   if (par->mDigit    <          0) par->mDigit = 0;
   if (par->mDigit    >          9) par->mDigit = 9;

	/* check if parameter match from readline */
	if (getType("COMMENT") - LINECOMMENT) 
		fprintf(stderr, "WARNING: readline mismatch - LINECOMMENT\n");
	if (getType("NULL")    - LINENULL)    
		fprintf(stderr, "WARNING: readline mismatch - LINENULL\n");
	if (getType("BLANK") - LINEBLANK) 
		fprintf(stderr, "WARNING: readline mismatch - LINEBLANK\n");
	if (getType("VALUE") - LINEVALUE) 
		fprintf(stderr, "WARNING: readline mismatch - LINEVALUE\n");
	if (getType("PAR") - LINEPAR) 
		fprintf(stderr, "WARNING: readline mismatch - LINEPAR\n");
	if (getType("UNKNOWN") - LINEUNKNOWN) 
		fprintf(stderr, "WARNING: readline mismatch - LINEUNKNOWN\n");

}




/******************************************************************************
	readInput: read input files and store in argv array

	usage: readInput(par, data)

	par    - parameters
	data   - information
******************************************************************************/
readInput(par, data)
parameters *par;
Data       *data;
{
	int n, maxLines;
	unsigned size;

	/* initialize */
	maxLines = FILELINES;
	n = 0;

	/* read input file and store */
	size = maxLines*sizeof(char *);
	data->fileLines = (char **) malloc(size);
	while((n += argvFile(data->fileLines + n, FILELINES, stdin)) >= maxLines) {
		maxLines += FILELINES;
		size = (unsigned) maxLines*sizeof(char *);
		data->fileLines = (char **) realloc((char *) data->fileLines, size);
		};
	par->n = n;

	/* null last pointer */
	data->fileLines[par->n] = NULL;

}




/******************************************************************************
	getInput: get input parameters and data (allocate memory)

	usage: getInput(par, data)

	par    - parameters
	data   - information
******************************************************************************/
getInput(par, data)
parameters  *par;
Data       *data;
{
	int lineType, b, f, g;
	char **line;

	/* set line types */
	setLine("fault ",      LINEFAULT);
   setLine("point ",      LINEPOINT);
   setLine("datum ",      LINEDATUM);
   setLine("block ",      LINEBLOCK);
   setLine("plane ",      LINEPLANE);
	setLine("options ",    LINEOPTION);
	setLine("stop ",       LINESTOP);

	/* count number of data items */
	par->b = 0; par->f = 0; par->g = 0;
	line = data->fileLines;
	while(*line != NULL) {
		switch((lineType = getLine(*line))) {
			case LINEFAULT:
				par->f++;
				break;
			case LINEDATUM:
			case LINEPLANE:
			case LINEBLOCK:
			case LINEPOINT:
				par->g++;
				setArg(*line);
				par->b += getNumberPoints(*line, lineType);
				endArg();
				break;
			default:
				break;
			}
		line++;
		}

	/* get memory */
	data->faults = (faultType    *) malloc(par->f*sizeof(faultType));
	data->groups = (PointGroup   *) malloc(par->g*sizeof(PointGroup)); 
	data->strain = (strainTensor *) malloc(par->b*sizeof(strainTensor));
	data->stress = data->strain;

	/* read in faults and point groups */
	f = 0; g = 0; b = 0;
	line = data->fileLines;
	while(*line != NULL) {
		switch(getLine(*line)) {
			case LINEFAULT:
				f += getFault(data->faults + f, *line);
				break;
			case LINEPOINT:
				setArg(*line);
				b += getPoint(data->strain + b, data->groups + g++, *line);
				endArg();
				break;
			case LINEDATUM:
				setArg(*line);
				b += getDatum(data->strain + b, data->groups + g++, *line);
				endArg();
				break;
			case LINEBLOCK:
				setArg(*line);
				b += getBlock(data->strain + b, data->groups + g++, *line);
				endArg();
				break;
			case LINEPLANE:
				setArg(*line);
				b += getPlane(data->strain + b, data->groups + g++, *line, par->XY);
				endArg();
				break;
			case LINEOPTION:
			case LINEPAR:
				setArg(*line);
				getOption(par, *line);
				endArg();
				break;
			case LINECOMMENT:
			case LINENULL:
			case LINEBLANK:
			case LINEVALUE:
			case LINEUNKNOWN:
			default:
				break;
			}
		line++;
		}

	/* end line types */
	endLine();

}




getNumberPoints(line, lineType)
char    *line;
int  lineType;
{
	int nx, ny, nz;
	switch(lineType) {
		case LINEPOINT:
			return(1);
			break;
		case LINEDATUM:
			mstArg("nx", "d", &nx);
			mstArg("ny", "d", &ny);
			return(nx*ny);
			break;
		case LINEBLOCK:
			mstArg("nx", "d", &nx);
			mstArg("ny", "d", &ny);
			mstArg("nz", "d", &nz);
			return(nx*ny*nz);
			break;
		case LINEPLANE:
			mstArg("nx", "d", &nx);
			mstArg("ny", "d", &ny);
			return(nx*ny);
			break;
		default:
			break;
		}
}




/******************************************************************************
	getFault: get fault

	usage: f = getFault(fault, line)

	fault  - fault information
	line   - input line

	return: f - number of faults (1)
******************************************************************************/
getFault(fault, line)
faultType *fault;   
char      *line;
{
   int input;
   char lineType[20], fType[20];
   input = sscanf(line, "%s %s %lf %lf %lf %lf %lf %lf %lf %lf %s %d %d",
                         lineType,     fType,
                        &fault->ds,   &fault->dl,   &fault->dt,   &fault->dw,
                        &fault->dp,   &fault->xLon, &fault->yLat, &fault->az,
                         fault->name, &fault->nx,   &fault->ny);
   fault->ft = *fType;
	fault->name[7] = '\0';
   if (input == 9) {
      fault->nx = 1;
      fault->ny = 1;
      }
   else if (input != 11) {
      fprintf(stderr, "ERROR: bad fault format\n");
      fprintf(stderr, "\"%s\"\n", line);
      exit(-1);
      }
	if (fault->ft != (char) 's' && fault->ft != (char) 'd' &&
	    fault->ft != (char) 'S' && fault->ft != (char) 'D') {
		fprintf(stderr, "ERROR: %c is not a valid fault type\n", fault->ft);
		exit(-1);
		}
	if (fault->dp > 90. || fault->dp <= 0.) {
		fprintf(stderr, "ERROR: dip (%f) must be > 0. and <= 90.\n", fault->dp);
		exit(-1);
		}
	return(1);
}




/******************************************************************************
	getPoint: get stress/strain point 

	usage: b = getPoint(strain, group, line)

	strain - strain information
	group  - point group
	line   - input line

	return: b - number of points in group (1)
******************************************************************************/
getPoint(strain, group, line)
strainTensor *strain;
PointGroup   *group;
char           *line;
{
	double x, y, z;
	int n;
	char name[20];

	/* initialize */
	strcpy(name, "point\0");
	z = 0.;
	n = 0;

	/* get parameters from line */
	mstArg("x",       "F",   &x   );
	mstArg("y",       "F",   &y   );
	getArg("z",       "F",   &z   );
	getArg("name,nm", "s",    name);

	/* get point group */
	group->line = strdup(line);
	strcpy(group->name, name);
	group->type = LINEPOINT;
	group->x1 = x; group->x2 = x;
	group->y1 = y; group->y2 = y;
	group->z1 = z; group->z2 = z;
	group->nx = group->ny = group->nz = 1;
	group->tensor = strain;

	/* generate point */
	strcpy(strain[n].name, name);
	strain[n].xLon = x;
	strain[n].yLat = y;
	strain[n].zEle = z;
	n++;

	/* return number of points */
	return(1);

}




/******************************************************************************
	getDatum: get stress/strain points on constant depth surface

	usage: b = getDatum(strain, group, line)

	strain - strain information
	group  - point group
	line   - input line

	return: b - number of points in group
******************************************************************************/
getDatum(strain, group, line)
strainTensor *strain;
PointGroup   *group;
char         *line;
{
	double x1, x2, y1, y2, z, dx, dy, angle, dip;
	int i, j, n, nx, ny;
	char name[20];

	/* initialize */
	strcpy(name, "datum\0");
	angle = 90.;
	dip = 90.;
	z = 0.;
	n = 0;


	/* get parameters from line */
	mstArg("x1",      "F",   &x1  );
	mstArg("x2",      "F",   &x2  );
	mstArg("y1",      "F",   &y1  );
	mstArg("y2",      "F",   &y2  );
	getArg("z",       "F",   &z   );
	mstArg("nx",      "d",   &nx  ); if (nx < 2) nx = 2;
	mstArg("ny",      "d",   &ny  ); if (ny < 2) ny = 2;
	getArg("name,nm", "s",    name); 
	getArg("angle,a", "F",   &angle);
	getArg("dip,d",   "F",   &dip);


	/* get point group */
	group->line = strdup(line);
	strcpy(group->name, name);
	group->type = LINEDATUM;
	group->x1 = x1; group->x2 = x2;
	group->y1 = y1; group->y2 = y2;
	group->z1 = z;  group->z2 = z;
	group->nx = nx; group->ny = ny; group->nz = 1;
	group->tensor = strain;
	group->angle  = angle; group->dip = dip;


	/* calculate limits */
	dx = (x2 - x1)/(nx - 1);
	dy = (y2 - y1)/(ny - 1);

	/* generate constant depth points row-by-row from (xmin, ymin) */
	for(j = 0; j < ny; j++) 
		for(i = 0; i < nx; i++) {
			strcpy(strain[n].name, name);
			strain[n].xLon = x1 + ((double) i)*dx;
			strain[n].yLat = y1 + ((double) j)*dy;
			strain[n].zEle = z; 
			n++;
			}

	/* return number of points */
	return(n);

}




/******************************************************************************
	getBlock: get stress/strain points in block

	usage: b = getBlock(strain, group, line)

	strain - strain information
	group  - point group
	line   - input line

	return: b - number of points in group
******************************************************************************/
getBlock(strain, group, line)
strainTensor *strain;
PointGroup   *group;
char         *line;
{
	double x1, x2, y1, y2, z1, z2, dx, dy, dz;
	int i, j, k, n, nx, ny, nz;
	char name[20];

	/* initialize */
	strcpy(name, "block\0");
	n = 0;

	/* get parameters from line */
	mstArg("x1",      "F", &x1  );
	mstArg("x2",      "F", &x2  );
	mstArg("y1",      "F", &y1  );
	mstArg("y2",      "F", &y2  );
	mstArg("z1",      "F", &z1  );
	mstArg("z2",      "F", &z2  );
	mstArg("nx",      "d", &nx  ); if (nx < 2) nx = 2;
	mstArg("ny",      "d", &ny  ); if (ny < 2) ny = 2;
	mstArg("nz",      "d", &nz  ); if (nz < 2) nz = 2;
	getArg("name,nm", "s",  name);

	/* get point group */
	group->line = strdup(line);
	strcpy(group->name, name);
	group->type = LINEBLOCK;
	group->x1 = x1; group->x2 = x2;
	group->y1 = y1; group->y2 = y2;
	group->z1 = z1; group->z2 = z2;
	group->nx = nx; group->ny = ny; group->nz = nz;
	group->tensor = strain;

	/* calculate limits */
	dx = (x2 - x1)/(nx - 1);
	dy = (y2 - y1)/(ny - 1);
	dz = (z2 - z1)/(nz - 1);

	/* generate block points row-by-row (from xmin, ymin, zmin) */
	for(k = 0; k < nz; k++)
		for(j = 0; j < ny; j++)
			for(i = 0; i < nx; i++) {
				strcpy(strain[n].name, name);
				strain[n].xLon = x1 + ((double) i)*dx;
				strain[n].yLat = y1 + ((double) j)*dy;
				strain[n].zEle = z1 + ((double) k)*dz;
				n++;
				}

	/* return number of points */
	return(n);

}




/******************************************************************************
	getPlane: get stress/strain points on plane

	usage: b = getPlane(strain, group, line, XY)

	strain - strain information
	group  - point group
	line   - input line
	XY     - data input mode (latitude/longitude or X/Y)

	return: b - number of points in group
******************************************************************************/
getPlane(strain, groups, line, XY)
strainTensor *strain;
PointGroup   *groups;
char           *line;
int               XY;
{
	faultType fault[1];
	double x0, y0, z0, sdx, sdy, sdz, ddx, ddy, ddz, zdx, zdy; 
	double nx, ny, xshift, yshift;
	int i, j, n;

	/* initialize */
	strcpy(fault->name, "plane\0");
	fault->dp = 90.;
	fault->dt = 0.;
	fault->az = 0.;
	n = 0;

	/* get parameters from line */
	mstArg("length,l",   "F",  &fault->dl  );
	getArg("depth,d",    "F",  &fault->dt  );
	mstArg("width,w",    "F",  &fault->dw  );
	getArg("dip,dp",     "F",  &fault->dp  );
	mstArg("x",          "F",  &fault->xLon);
	mstArg("y",          "F",  &fault->yLat);
	getArg("azimuth,az", "F",  &fault->az  );
	mstArg("nx",         "d",  &fault->nx  );
	mstArg("ny",         "d",  &fault->ny  );
	getArg("name,nm",    "s",   fault->name); 

	/* get point group */
	groups->line = strdup(line);
	strcpy(groups->name, fault->name);
	groups->type = LINEPLANE;
	groups->x1 = fault->xLon; groups->x2 = fault->xLon;
	groups->y1 = fault->yLat; groups->y2 = fault->yLat;
	groups->z1 = fault->dt  ; groups->z2 = fault->dt  ;
	groups->dl = fault->dl  ; groups->dw = fault->dw  ;
	groups->az = fault->az  ; groups->dp = fault->dp  ;
	groups->nx = fault->nx  ; groups->ny = fault->ny  ; groups->nz = 1;
	groups->angle = fault->az; groups->dip = fault->dp;
	groups->tensor = strain;
 
	/* get grid placement parameters nx and ny */
	nx = (double) ((fault->nx > 0) ? (fault->nx - 1) : (abs(fault->nx)));
	ny = (double) ((fault->ny > 0) ? (fault->ny - 1) : (abs(fault->ny)));
	xshift = (fault->nx > 0) ? (0.) : (0.5);
	yshift = (fault->ny > 0) ? (0.) : (0.5);

	/* get subfault increments due to strike */
	sdx = fault->dl*cos(fault->az/(RAD))/nx;
	sdy = fault->dl*sin(fault->az/(RAD))/nx;
 
	/* get subfault increments due to dip */
	ddx = fault->dw*cos(fault->dp/(RAD))*cos((fault->az + 90.)/(RAD))/ny;
	ddy = fault->dw*cos(fault->dp/(RAD))*sin((fault->az + 90.)/(RAD))/ny;
	ddz = fault->dw*sin(fault->dp/(RAD))/ny;
 
	/* get initial increment due to fault depth (not for 0 dip) */
	zdx = zdy = fault->dt*cos(fault->dp/(RAD))/sin(fault->dp/(RAD));
	zdx = zdx*cos((fault->az + 90.)/(RAD)); 
	zdy = zdy*sin((fault->az + 90.)/(RAD)); 
 
	/* convert into units of latitude and longitude */
	if (!XY) {
		sdx = sdx/(KMD)/cos(fault->yLat/(RAD));
		sdy = sdy/(KMD);
		ddx = ddx/(KMD)/cos(fault->yLat/(RAD));
		ddy = ddy/(KMD);
		zdx = zdx/(KMD)/cos(fault->yLat/(RAD));
		zdy = zdy/(KMD);
		}

	/* calculate starting point (lower-left) */
	x0 = fault->xLon + zdx + (ny - yshift)*ddx - (nx/2. - xshift)*sdx;
	y0 = fault->yLat + zdy + (ny - yshift)*ddy - (nx/2. - xshift)*sdy;
	z0 = fault->dt +         (ny - yshift)*ddz;
 
	/* generate plane points row-by-row from lower-left */
	for(j = 0; j < abs(fault->ny); j++) {
		for(i = 0; i < abs(fault->nx); i++) {
			strcpy(strain[n].name, fault->name);
			strain[n].xLon = x0 + ((double) i)*sdx - ((double) j)*ddx;
			strain[n].yLat = y0 + ((double) i)*sdy - ((double) j)*ddy;
			strain[n].zEle = z0                    - ((double) j)*ddz;
			n++;
			}
		}
 
	/* return number of points */
	return(n);

}




/******************************************************************************
	getOption: get options

	usage: getOption(par, line)

	par    - parameters
	line   - input line
******************************************************************************/
getOption(par, line)
parameters  *par;
char       *line;
{
	getArg("xy"    ,    "d", &par->XY); 
	getArg("stress",    "d", &par->stress);
	getArg("mdigit",    "d", &par->mDigit);
	getArg("increment", "F", &par->increment);
	getArg("faultfile", "s",  par->faultFile);
}




/******************************************************************************
	getStrain: get strain tensor

	usage: getStrain(par, strain, faults)

	par    - parameters
	strain - strain information
	faults - fault parameters
******************************************************************************/
getStrain(par, strain, faults)
parameters *par;
strainTensor *strain;
faultType  *faults;
{
	benchMark point[6];
	double Ux, Uy, Uz;
	int i, j, k, XY;

	/* initialize */
	XY = par->XY;

	/* loop over points and faults */
	for(i = 0; i < par->b; i++) {
		getPoints(point, strain + i, par->increment, XY);
		for(j = 0; j < par->f; j++) {
			for(k = 0; k < 6; k++) {
				getDisl(&Ux, &Uy, &Uz, &(point[k]), faults + j, XY); 
				point[k].xMod += Ux*faults[j].ds;
				point[k].yMod += Uy*faults[j].ds;
				point[k].zMod += Uz*faults[j].ds;
				}
			}
		getStrainTensor(strain + i, point, par->increment);
		}

}




getPoints(point, strain, increment, XY)
benchMark    *point;
strainTensor *strain;
double        increment;
int           XY;
{
	
	double dx, dy, dz;
	int i;

	/* get increments */
	dx = increment; dy = increment; dz = increment;
	if (!XY) {
		dx = increment/((KMD)*cos(strain->yLat/(RAD)));
		dy = increment/(KMD);
		}

	/* get initial points */
	for(i = 0; i < 6; i++) {
		point[i].xLon = strain->xLon;
		point[i].yLat = strain->yLat;
		point[i].zEle = strain->zEle;
		point[i].xMod = 0.;
		point[i].yMod = 0.;
		point[i].zMod = 0.;
		}

	/* get new points */
	point[0].xLon += dx;
	point[1].xLon -= dx;
	point[2].yLat += dy;
	point[3].yLat -= dy;
	point[4].zEle += dz;
	point[5].zEle -= dz;
	if (point[5].zEle < 0.) point[5].zEle = 0.;

}




getDisl(ux, uy, uz, point, f, XY)
benchMark *point;
faultType *f;
double *ux, *uy, *uz;
int XY;
{
	double fx, fy, fz, x, y, z, cs, sn;

	/* initialize */
	cs = cos(f->az/(RAD));
	sn = sin(f->az/(RAD));

	/* change from Cartesian to fault coordinates */
	fx = point->xLon - f->xLon;
	fy = point->yLat - f->yLat;
	if (!XY) {
		fx = fx*(KMD)*cos(f->yLat/(RAD));
		fy = fy*(KMD);
		}
	x = cs*fx + sn*fy;
	y = cs*fy - sn*fx;
	z = fabs(point->zEle);
   if (f->ft == (char) 'S' || f->ft == (char) 'D') {
      if (f->dp <= 89.) y += f->dt/tan(f->dp/(RAD));
      }

	/* get displacements in fault coordinate system */
	getdisl(x, y, z, f->dl, f->dt, f->dw, f->dp, f->ft, &fx, &fy, &fz);

	/* rotate displacements from fault to Cartesian coordinates */
	cs = cos(f->az/(RAD));
	sn = sin(f->az/(RAD));
	*ux = fx*cs - fy*sn;
	*uy = fx*sn + fy*cs;
	*uz = fz;
	
}




getStrainTensor(strain, point, increment)
strainTensor *strain;
benchMark    *point;
double        increment;
{
	double dx, dy, dz;

	/* get increments */
	dx = 2.*increment;
	dy = 2.*increment;
	dz = 2.*increment;
	if ((fabs(point->zEle)-increment) < 0.) dz = increment + fabs(point->zEle); 

	/* change from km to cm */
	dx *= CMKM;
	dy *= CMKM;
	dz *= CMKM;

	/* get compressional strains (Txx, Tyy, Tzz) */
	strain->Txx = (point[0].xMod - point[1].xMod)/dx;
	strain->Tyy = (point[2].yMod - point[3].yMod)/dy;
	strain->Tzz = (point[5].zMod - point[4].zMod)/dz;

	/* get shear strains */ 
	strain->Txy = 0.5*((point[0].yMod - point[1].yMod)/dx + 
	                   (point[2].xMod - point[3].xMod)/dy);
	strain->Txz = 0.5*((point[0].zMod - point[1].zMod)/dx + 
	                   (point[5].xMod - point[4].xMod)/dz);
	strain->Tyz = 0.5*((point[2].zMod - point[3].zMod)/dy + 
	                   (point[5].yMod - point[4].yMod)/dz);

}





/******************************************************************************
	getMicroStrain: convert strain to microstrain

	usage: getMicroStrain(par, strain)

	par    - parameters
	strain - strain information
******************************************************************************/
getMicroStrain(par, strain)
parameters   *par;
strainTensor *strain;
{
	double microstrain;
	int b;
	microstrain = 1./(MICRO);
	for(b = 0; b < par->b; b++) {
		strain[b].Txx *= microstrain; 
		strain[b].Tyy *= microstrain; 
		strain[b].Tzz *= microstrain; 
		strain[b].Txy *= microstrain; 
		strain[b].Txz *= microstrain; 
		strain[b].Tyz *= microstrain; 
		}
}




/******************************************************************************
	getStress: convert strain tensor to stress tensor

	usage: getStress(par, stress)

	par    - parameters
	stress - stress information
******************************************************************************/
getStress(par, stress)
parameters   *par;
stressTensor *stress;
{
	strainTensor *strain;
	double lamda, mu2, dilatation;
	int b;

	/* initialize */
	strain =    stress;
	lamda  =    par->lamda;
	mu2    = 2.*par->mu;

	/* get stress tensor for all points */
	for(b = 0; b < par->b; b++) {
		dilatation = lamda*(strain[b].Txx + strain[b].Tyy + strain[b].Tzz);
		stress[b].Txx = dilatation + mu2*strain[b].Txx;
		stress[b].Tyy = dilatation + mu2*strain[b].Tyy;
		stress[b].Tzz = dilatation + mu2*strain[b].Tzz;
		stress[b].Txy = mu2*strain[b].Txy;
		stress[b].Txz = mu2*strain[b].Txz;
		stress[b].Tyz = mu2*strain[b].Tyz;
		}

}




/******************************************************************************
	getFaultStress: get stress components at orientation on plane or datum

	usage: getFaultStress(par, data)

	par    - parameters
	data   - input data

	note: Cr > 0 is earthquake inducing (right-lateral)
	note: Cl < 0 is earthquake inducing (left- lateral) 
	note: Cn > 0 is earthquake inducing (normal)
	note: Ct < 0 is earthquake inducing (thrust)
******************************************************************************/
getFaultStress(par, data)
parameters   *par;
Data        *data;
{
	stressTensor *stress;
	PointGroup   *groups;
	double Nn[3], Ns[3], Nd[3], Ts, Td, Tn, Cr, Cl, Cn, Ct;
	int i, j, n;

	for(i = 0; i < par->g; i++) {
		switch(data->groups[i].type) {
			case LINEPLANE:
			case LINEDATUM:
				groups = data->groups + i;
				stress = groups->tensor;
				getPlaneNormal(groups->angle, groups->dip, Nn, Ns, Nd);
				n = groups->nx*groups->ny;
				for(j = 0; j < n; j++) {
					getPlaneStress(stress + j, Nn, Ns, Nd, &Ts, &Td, &Tn);
					Cr = Ts + par->friction*Tn; /* > 0: earthquake inducing */
					Cl = Ts - par->friction*Tn; /* < 0: earthquake inducing */
					Cn = Td + par->friction*Tn; /* > 0: earthquake inducing */
					Ct = Td - par->friction*Tn; /* < 0: earthquake inducing */
					stress[j].Txx = Ts; stress[j].Tyy = Td; stress[j].Tzz = Tn;
					stress[j].Txy = Cr; stress[j].Txz = Cl; stress[j].Tyz = 0.;
					}
				break;
			default:
				break;
			}
		}

}




/******************************************************************************
	getPlaneNormal: determine normal vectors to plane

	usage: getPlaneNormal(azimuth, dip, Nn, Ns, Nd)

	azimuth - strike of plane orientation (clockwise from east)
	dip     - to north for 0 azimuth (to south for 180 azimuth)
	Nn      - normal to plane (outward if dipping toward you)
	Ns      - normal along strike (goes to left)
	Nd      - normal along dip (down if dipping toward you)

	note: normal vectors defined where N[0]=east, N[1]=north, N[2]=up
******************************************************************************/
getPlaneNormal(azimuth, dip, Nn, Ns, Nd)
double azimuth, dip, *Nn, *Ns, *Nd;
{
	azimuth =        azimuth/(RAD);
	dip     =            dip/(RAD);
	Nn[0] = -sin(dip)*sin(azimuth);
	Nn[1] =  sin(dip)*cos(azimuth);
	Nn[2] =  cos(dip)             ;
	Ns[0] =           cos(azimuth);
	Ns[1] =           sin(azimuth);
	Ns[2] =                     0.;
	Nd[0] = -cos(dip)*sin(azimuth);
	Nd[1] =  cos(dip)*sin(azimuth);
	Nd[2] = -sin(dip)             ;
}





/******************************************************************************
	getPlaneStress: use stress tensor to get stresses at specified orientation

	usage: getPlaneStress(stress, Nn, Ns, Nd, &sshear, &dshear, &normal)

	stress  - stress tensor information (or strain)
	Nn      - normal to plane (outward if dipping toward you)
	Ns      - normal along strike (goes to left)
	Nd      - normal along dip (down if dipping toward you)
	sshear  - (return) stike shear (right-lateral is positive)
	dshear  - (return) dipping shear (down-dip is positive) 
	normal  - (return) normal stress (tension is positive)

	note: normal vectors defined where N[0]=east, N[1]=north, N[2]=up
******************************************************************************/
getPlaneStress(stress, Nn, Ns, Nd, sshear, dshear, normal)
stressTensor *stress;
double *Nn, *Ns, *Nd, *sshear, *dshear, *normal;
{
	double T[3]; 

	/* determine traction vector */
	T[0] = stress->Txx*Nn[0] + stress->Txy*Nn[1] + stress->Txz*Nn[2];
	T[1] = stress->Txy*Nn[0] + stress->Tyy*Nn[1] + stress->Tyz*Nn[2];
	T[2] = stress->Txz*Nn[0] + stress->Tyz*Nn[1] + stress->Tzz*Nn[2];

	/* determine fault stresses */
	*normal = T[0]*Nn[0] + T[1]*Nn[1] + T[2]*Nn[2];
	*sshear = T[0]*Ns[0] + T[1]*Ns[1] + T[2]*Ns[2];
	*dshear = T[0]*Nd[0] + T[1]*Nd[1] + T[2]*Nd[2];
	
}




/******************************************************************************
	writeOutput: write results to standard output

	usage: writeOutput(par, data)

	par    - parameters
	data   - input data
******************************************************************************/
writeOutput(par, data)
parameters   *par;
Data        *data;
{
	strainTensor *strain;
	stressTensor *stress;
	PointGroup   *groups;
	faultType         *f;
	int          i, j, n;
	char      format[82];

	/* initialize */
	groups = data->groups;
	f      = data->faults;

   /* write header and input parameters */
   printf("                              STRESS/STRAIN MODELING\n");
   printf("                              ----------------------\n");
   printf("options xy=%-d stress=%-d mdigit=%-d increment=%-8.5lf\n", 
			par->XY, par->stress, par->mDigit, par->increment);

	/* write header information for faults */
	printFLabel(par->XY);

	/* write faults */
   for(i = 0; i < par->f; i++) {
      printf("fault ");
      printf("%c %8.2lf %7.2lf %6.2lf %6.2lf %5.1lf %9.4lf %8.4lf %6.1lf  ",
              f[i].ft, f[i].ds, f[i].dl, f[i].dt, f[i].dw, f[i].dp,
              f[i].xLon, f[i].yLat, f[i].az);
      printf("%-7s",  f[i].name);
      printf("\n");
      }

	/* write header information for strain/stress tensor */
	printf("\n");
	printf("\n");
	if (!par->stress) {
		printf("                                 MICROSTRAIN TENSOR\n");
		printf("                                 ------------------\n");
		}
	else {
		printf("                                STRESS TENSOR (BARS)\n");
		printf("                                -------------------\n");
		}

	/* get default format string */
   sprintf(format, "%%9. lf%%8. lf%%7. lf%%+7s%%8. lf%%8. lf%%8. lf");
   sprintf(format + 40, "%%8. lf%%8. lf%%8. lf\n\0");

	/* write stress/strain tensor for points */
	n = 0;
	for(i = 0; i < par->g; i++) {
		if (groups[i].type == LINEPOINT) {
			if (n++ == 0) printSLegend(par, NULL,  LINEPOINT);
			stress = groups[i].tensor;
			printPoint(format, par, stress);
			}
		}

	/* write stress/strain tensor for point groups */
	for(i = 0; i < par->g; i++) {
		switch(groups[i].type) {
			case LINEDATUM:
			case LINEBLOCK:
			case LINEPLANE:
				printSLegend(par, groups[i].line, groups[i].type);
				stress = groups[i].tensor;
				n = groups[i].nx*groups[i].ny;
				if (groups[i].type == LINEBLOCK) n *= groups[i].nz;
				for(j = 0; j < n; j++) printPoint(format, par, stress + j);
				break;
			default:
				break;
			}

		}

}




printFLabel(XY) 
int XY;
{
	printf("\n");
	printf("\n");
	printf("                                  FAULTS MODELED\n");
	printf("                                  --------------\n");
	printf("\n");
   if (XY) {
      printf("    Ftype   Slip  Length  Depth  Width   Dip ");
      printf("    XDist    YDist Azimuth Name\n");
      printf("    -----   ----  ------  -----  -----   --- ");
      printf("    -----    ----- ------- ----\n");
      }
   else {
      printf("    Ftype  Slip   Length  Depth  Width   Dip ");
      printf("Longitude Latitude Azimuth Name\n");
      printf("    -----  ----   ------  -----  -----   --- ");
      printf("--------- -------- ------- ----\n");
      }
}




printSLegend(par, line, mode)
parameters  *par;
int         mode;
char       *line;
{
   int n = 0;
   char string[258];
   switch(mode) {
      case LINEPOINT:
         printf("\n");
         printf("POINT\n");
         printf("\n");
         printSLabel(par);
         break;
		case LINEDATUM:
         while(line[n] == (char) ' ') n++;
         printf("\n\n");
         strcpy(string, line + n);
         strncpy(string, "DATUM", 5);
         printf("%s\n", string);
         printf("\n");
         printSLabel(par);
         break;
		case LINEBLOCK:
         while(line[n] == (char) ' ') n++;
         printf("\n\n");
         strcpy(string, line + n);
         strncpy(string, "BLOCK", 5);
         printf("%s\n", string);
         printf("\n");
         printSLabel(par);
         break;
		case LINEPLANE:
         while(line[n] == (char) ' ') n++;
         printf("\n\n");
         strcpy(string, line + n);
         strncpy(string, "PLANE", 5);
         printf("%s\n", string);
         printf("\n");
         printSLabel(par);
         break;
      }
}




printSLabel(par)
parameters *par;
{
	if (par->XY == 1)   printf("    XDIST   YDIST  DEPTH   NAME     ");
	if (par->XY != 1)   printf("LONGITUDE LATITUDE DEPTH   NAME     ");
	if (par->stress==2) printf(" Ts      Td      Tn      Cr     Cl         \n");
	if (par->stress!=2) printf("Txx     Tyy     Tzz     Txy     Txz     Tyz\n");
	if (par->XY == 1)   printf("    -----   -----  -----   ----     ");
	if (par->XY != 1)   printf("--------- --------  ----   ----     ");
	if (par->stress==2) printf("---     ---     ---     ---     ---     ---\n");
	if (par->stress!=2) printf("---     ---     ---     ---     ---     ---\n");
}




printPoint(format, par, stress)
parameters      *par;
stressTensor *stress;
char         *format;
{
	getFormat(format, par, stress);
	if (par->stress == 2) {
		printf(format, stress->xLon, stress->yLat, stress->zEle, 
		               stress->name,
		               stress->Txx , stress->Tyy , stress->Tzz ,
		               stress->Txy , stress->Txz , stress->Tyz );
		}
	else {
		printf(format, stress->xLon, stress->yLat, stress->zEle, 
		               stress->name,
		               stress->Txx , stress->Tyy , stress->Tzz ,
		               stress->Txy , stress->Txz , stress->Tyz );
		}
}




getFormat(format, par, stress)
parameters   *par;
stressTensor *stress;
char         *format;
{
	format[ 3] = (char) (getMDigit(stress->xLon, 9,           4) + '0');
	format[ 9] = (char) (getMDigit(stress->yLat, 8,           4) + '0');
	format[15] = (char) (getMDigit(stress->zEle, 7,           2) + '0');
	format[25] = (char) (getMDigit(stress->Txx,  8, par->mDigit) + '0');
	format[31] = (char) (getMDigit(stress->Tyy,  8, par->mDigit) + '0');
	format[37] = (char) (getMDigit(stress->Tzz,  8, par->mDigit) + '0');
	format[43] = (char) (getMDigit(stress->Txy,  8, par->mDigit) + '0');
	format[49] = (char) (getMDigit(stress->Txz,  8, par->mDigit) + '0');
	format[55] = (char) (getMDigit(stress->Tyz,  8, par->mDigit) + '0');
}




int
getMDigit(value, field, mDigit)
double value;
int    field, mDigit;
{
	int max, n;
	char fTest[20], mTest[20];
	sprintf(fTest, "%%%d.%dlf\0", field, mDigit);
	sprintf(mTest, fTest, value);
	max = (mTest[0] == (char) '-' || mTest[0] == ' ') ? (field) : (field - 1);
	if ((n = strlen(mTest)) > max) if ((mDigit -= n - max) < 0) mDigit = 0;
	return(mDigit);
}
