/******************************************************************************
	dist: time-dependent displacement modeling by dislocations 

    ================================================================= 
   | (c) Copyright 1998 The Regents of the University of California. |
   |     All Rights Reserved.                                        |
   |                                                                 |
   |     This work was produced under the sponsorship of the         |
   |     United States Department of Energy.                         |
   |     The Government retains certain rights therein.              |
    ================================================================= 

	usage:   dist [optional parameters] < inputFile > displacementFile

	compile: cc -g -o dist dist.c getdisl.c readinput.a -lm -lget

	dist computes the time-dependent vertical and horizontal deformations 
	at longitude and latitude points (benchmarks) on the earth
	using the dislocation models of chinnery, BSSA 1961 (#3), 
	savage and hastie, JGR 1966 (#20), mansinha and smylie, JGR 1967 (#18) 
	and mansinha and smylie, BSSA 1971, (#5).
	a file containing the name of all parameters, faults, and benchmarks
	is used as input although separate fault and data files can be used.
	points can be entered as longitude/latitude/elevation or x/y/z.

	author: shawn larsen
	date:   january 1998
	where:  llnl
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "disl.h"
#include "usage.h"
#include "sacutil.h"

#define      DT    0.01
#define       N   10000
#define       V     4.0
#define     MMU  2.8e11
#define  STRIKE   124.0
#define     DIP    90.0
#define  SACLAT    37.797
#define  SACLON  -122.380
#define MFACTOR  1.0e21
#define    TYPE     's'

main(argc, argv)
int    argc;
char **argv;
{
	double lon, lat, depth, width, length, slip, rise, strike, dip, time, moment;
	double mu, mfactor, v, distance, start, end, f, finc;
	double saclon, saclat, x, y, z, ux, uy, uz;
	float  *sacX, *sacY, *sacZ, dt;
	int i, n, istart, iend;
	char line[82], type;
	FILE *file;

	/* get memory and initialize */
	n       = N;
	dt      = DT;
	mu      = MMU;
	saclon  = SACLON;
	saclat  = SACLAT;
	strike  = STRIKE;
	dip     = DIP;
	mfactor = MFACTOR;
	v       = V;
	type    = (char) TYPE;
	sacX = (float *) malloc(n*sizeof(float));
	sacY = (float *) malloc(n*sizeof(float));
	sacZ = (float *) malloc(n*sizeof(float));
	for(i = 0; i < n; i++) {
		sacX[i] = 0.;
		sacY[i] = 0.;
		sacZ[i] = 0.;
		}

	/* loop over stations */

	/* loop over fault elements */
	while(fgets(line, 82, stdin) != NULL) {
		i = 0;
		for(i = 0; i < 82; i++) {
			if (line[i] == (char) '\0') break;
			if (line[i] == (char)  'E') line[i] = (char) 'e';
			}
		sscanf(line, "%lf %lf %lf %lf %lf %lf %lf",
		       &lat, &lon, &depth, &slip, &rise, &time, &moment);
		length = (sqrt((moment*mfactor)/(mu*slip)))/(CMKM);
		width  = length;
		getFaultCoord(&x, &y, saclon, saclat, lon, lat, strike, dip, depth, 0);
		z = 0.;
		getDisl(&ux, &uy, &uz, x, y, strike, dip, length, depth, width, type);
		ux *= slip;
		uy *= slip;
		uz *= slip;
		distance = sqrt(x*x + y*y);
		start = distance/v + time;
		end   = start + rise;
		istart = (int) (start/((double) dt));
		iend   = (int) (  end/((double) dt));
		if (istart < 0) istart == 0; 
		if (iend   < 0) iend   == 0; 
		if (istart > n) istart == n;
		if (iend   > n) iend   == n;
		if (istart > iend) iend == istart;
		finc = (((double) dt)/rise);
		f = 0.;
		i = istart;
		while(i <= iend) {
			sacX[i] += (float) f*ux;
			sacY[i] += (float) f*uy;
			sacZ[i] += (float) f*uz;
			f += finc;
			i++;
			}
		i = iend + 1;
		while(i < n) {
			sacX[i] += (float) ux;
			sacY[i] += (float) uy;
			sacZ[i] += (float) uz;
			i++;
			}
		}
	
	/* write out sac files */
	file = fopen("out.X", "w");
	writeSAC(sacX, n, dt, file);
	fclose(file);
	file = fopen("out.Y", "w");
	writeSAC(sacY, n, dt, file);
	fclose(file);
	file = fopen("out.Z", "w");
	writeSAC(sacZ, n, dt, file);
	fclose(file);

	/* free memory */
	free(sacX);
	free(sacY);
	free(sacZ);

}


getFaultCoord(x, y, saclon, saclat, lon, lat, strike, dip, depth, xy)
double *x, *y, saclon, saclat, lon, lat, strike, dip, depth;
int xy;
{
	double cs, sn, dx, dy;
	cs = cos((strike)/(RAD));
	sn = sin((strike)/(RAD));
	dx = saclon - lon;
	dy = saclat - lat;
	if (!xy) {
		dx = dx*(KMD)*cos(lat/(RAD));
		dy = dy*(KMD);
		}
	*x = cs*dx + sn*dy;
	*y = cs*dy - sn*dx;
	if (dip <= 89.) *y += depth/tan(dip/(RAD));
}


getDisl(ux, uy, uz, x, y, strike, dip, length, depth, width, type)
double strike, dip, length, depth, width;
double *ux, *uy, *uz, x, y;
char type;
{
	double fx, fy, z, cs, sn;

	/* get displacements at surface (depth = 0) */
	z = 0.;

	/* get displacements in fault coordinate system */
	getdisl(x, y, z, length, depth, width, dip, type, &fx, &fy, uz);

	/* rotate displacements from fault to Cartesian coordinates */
/*
	cs = cos(strike/(RAD));
	sn = sin(strike/(RAD));
	*ux = fx*cs - fy*sn;
	*uy = fx*sn + fy*cs;
*/
*ux = fx;
*uy = fy;
	
}



writeSAC(array, n, dt, file)
float *array, dt;
int n;
FILE *file;
{
   SACParameter sacParameter[1];
   char sacHeader[SACSIZE];

   /* get sacFile header */
   getSACDefault(sacParameter);
   sacParameter->delta  = dt;
   sacParameter->npts   = n;
   sacParameter->b      = 0.;
   sacParameter->e      = (float) (n - 1)*dt;
   sacParameter->iftype = 1;
   sacParameter->leven  = 1;
   getSACHeader(sacHeader, sacParameter);

   /* write sac file */
   fwrite(sacHeader, SACSIZE, 1, file);
   fwrite(array, sizeof(float), n, file);

}

