/******************************************************************************
	gpsinv: 3D geodetic inversion for the slip distribution on faults

	gpsinv determines the slip distribution on arbitrary fault planes
	for a 3D geodetic network geometry.
	the model assumes an overdetermined system and includes constraints
	for surface slip distribution and smoothness.

	usage: gpsinv par=parameter_file > output_file

	compile: cc -g -o gpsinv gpsinv.c -lm -lget

	author: shawn larsen
	date:   january 1989
	where:  caltech
******************************************************************************/

#define BENCHMARKS 20000
#define FAULTS       100

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "disl.h"
#include "usage.h"

typedef struct parameter_gpsinv parameters;

main(argc, argv)
int    argc;
char **argv;
{
	parameters *par;
	benchMark  *bm;
	faultType  *faults;
	double     **A, *B, **C, *Slip;
	int i, m, n;

	/* print usage if no parameters */
	if (argc == 1) exit(getUsage());


	/* initialize */
	par = (parameters *) malloc(sizeof(parameters));


	/* get parameters */
	setpar(argc, argv);
	getPar(par);
	endpar();


	/* get benchmark and fault memory */
	bm     = (benchMark *) malloc(par->b*sizeof(benchMark));
	faults = (faultType *) malloc(par->f*sizeof(faultType));
	Slip   = (double    *) malloc(par->f*sizeof(double   ));


	/* read benchmark file */
	readData(bm, par);


	/* read fault file */
	readFault(faults, par);


	/* get matrix and data memory */
	m = par->m;
	n = par->n;
	if (par->s) m += par->n;
	if (par->a) m += par->n;
	if (par->d) m += par->f;
	A = (double **) malloc(m*sizeof(double *));
	for(i = 0; i < m; i++) A[i] = (double *) malloc(m*sizeof(double));
	B = (double  *) malloc(m*sizeof(double  ));
	C = (double **) malloc(n*sizeof(double *));
	for(i = 0; i < n; i++) C[i] = (double    *) malloc(n*sizeof(double));


	/* form dislocation kernals (A matrix) */
	getKernal(A, par, bm, faults);


	/* get data vector */
	getData(B, par, bm);


	/* normalize kernal and data by measurement errors */
	getUnitError(A, B, par, bm);


	/* add normalized a priori slip contraints */
	if (par->a) par->m += getASlip(A, B, par, faults);


	/* add normalized smoothness constraints */
	if (par->s) par->m += getSmoothness(A, B, par, faults);


	/* add normalized moment constraints */
	if (par->d) par->m += getMoment(A, B, par, faults);


	/* invert kernal to get covariance */
	getCov(C, A, par);


	/* calculate subfault slip */
	getSlip(Slip, par, A, B, C);


	/* write output */
	writeOutput(par, bm, faults, C, Slip);


	/* free memory */
	for(i = 0; i < par->m; i++) free(A[i]); free(A);
	for(i = 0; i < par->n; i++) free(C[i]); free(C);
	free(B);
	if (par->f) free(par->mError);
	free(faults);
	free(Slip);
	free(bm);
	free(par);
	
}



/******************************************************************************
	getUsage: print usage if no parameters

	usage: getUsage()
******************************************************************************/
getUsage()
{

	/* print usage message to stderr */
	fprintf(stderr, "Usage: gpsinv [parameters] [par=parFile] > outputFile\n");

	/* print parameters to stdout */
	printf("benchmarks=100\n");
	printf("faults=20\n");
	printf("xy=0\n");
	printf("xerror=1.0\n");
	printf("yerror=1.0\n");
	printf("zerror=1.0\n");
	printf("datafile=\"datafile\"\n");
	printf("faultfile=\"faultfile\"\n");
	printf("smoothnesserror=0.\n");
	printf("sliperror=0.\n");
	printf("slipfile=\"\"\n");
	printf("momenterror=0.,0.\n");

	/* return exit value */
	return(-1);

}



/******************************************************************************
	getPar: read program parameters

	usage: getPar(par)

	par - parameters
******************************************************************************/
getPar(par)
parameters *par;
{
	int i, d;
	char Fsize[10];

	/* initialize parameters */
	par->b           = BENCHMARKS;
	par->f           = FAULTS;
	par->XY          = XYFLAG;
	par->aError      = AERROR * (double)   SLIPFLAG;
	par->sError      = SERROR * (double) SMOOTHFLAG;
	par->slipFile[0] = (char) '\0';
	par->a           = 0;
	par->d           = 0;
	par->s           = 0;

	/* get parameters */
	getpar("benchmarks"      , "d", &par->b);
	getpar("faults"          , "d", &par->f);
	getpar("xy"              , "d", &par->XY); 
	getpar("sliperror"       , "F", &par->aError);
	getpar("smoothnesserror" , "F", &par->sError);
	getpar("xerror"          , "F", &par->xError);
	getpar("yerror"          , "F", &par->yError);
	getpar("zerror"          , "F", &par->zError);
	mstpar("datafile"        , "s",  par->dataFile);
	mstpar("faultfile"       , "s",  par->faultFile);
	getpar("slipfile"        , "s",  par->slipFile);

	/* get memory and parameters for moment */
	par->mError = (double *) malloc(par->f*sizeof(double));
	for(i = 0; i < par->f; i++) par->mError[i] = MERROR * (double) MOMENTFLAG;
	sprintf(Fsize, "vF[%d]\0", par->f);
	d = getpar("momenterror" , Fsize,  par->mError);
	for(i = 0; i < d; i++) if (par->mError[i] > 0.) par->d++;

	/* check parameters */
	if (par->b      < BENCHMARKS) par->b = BENCHMARKS;
	if (par->f      <     FAULTS) par->f = FAULTS;
	if (par->aError <         0.) par->aError = AERROR;
	if (par->aError >         0 ) par->a = 1;
	if (par->slipFile[0]        ) par->a = 1;
	if (par->sError <         0.) par->sError = SERROR;
	if (par->sError >         0.) par->s = 1;
	if (par->xError <         0.) par->xError = 0.;
	if (par->yError <         0.) par->yError = 0.;
	if (par->zError <         0.) par->zError = 0.;
	for(i = 0; i < par->f; i++) if (par->mError[i] < 0.) par->mError[i] = MERROR;

}




/******************************************************************************
	readData: read data file of benchmark locations and error (errors optional)

	usage: readData(bm, par)

	bm  - benchmark information
	par - parameters
******************************************************************************/
readData(bm, par)
benchMark  *bm;
parameters *par;
{
	double xError, yError, zError;
	int e, n;
	char line[82];
	FILE *in;

	/* initialize */
	n = 0;

	/* open data file  */
	in = fopen(par->dataFile, "r");
   if (in == NULL) {
      fprintf(stderr, "ERROR: can't open data file %s\n", par->dataFile);
      exit(-1);
      }

	/* set default error for first benchmark */
	bm[0].xErr = XERROR; bm[0].yErr = YERROR; bm[0].zErr = ZERROR;
	if (par->xError > 0.) bm[0].xErr = par->xError;
	if (par->yError > 0.) bm[0].yErr = par->yError;
	if (par->zError > 0.) bm[0].zErr = par->zError;

	/* read file line by line, skipping comments */
	while(fgets(line, 82, in)) {
		/*if (commentLine(line, 82)) continue;*/
		if (getLine(line) != LINEDATA) continue;
		strncpy(bm[n].name, line + 24, 7); bm[n].name[7] = (char) '\0';
		sscanf(line +  0, "%lf %lf %lf", &bm[n].xLon, &bm[n].yLat, &bm[n].zEle);
		sscanf(line + 31, "%lf %lf %lf", &bm[n].xDat, &bm[n].yDat, &bm[n].zDat);
   	if (line[17] != (char) ' ') sscanf(line + 17, "%lf", &bm->zEle);
		e = sscanf(line + 55, "%lf %lf %lf", &xError, &yError, &zError);
		if (par->xError > 0)
			bm[n].xErr = par->xError;
		else if (e == 3 && xError > 0.)
			bm[n].xErr = xError;
		else if (n)
			bm[n].xErr = bm[n - 1].xErr;
		if (par->yError > 0)
			bm[n].yErr = par->yError;
		else if (e == 3 && zError > 0.)
			bm[n].yErr = yError;
		else if (n)
			bm[n].yErr = bm[n - 1].yErr;
		if (par->zError > 0)
			bm[n].zErr = par->zError;
		else if (e == 3 && zError > 0.)
			bm[n].zErr = zError;
		else if (n)
			bm[n].zErr = bm[n - 1].zErr;
		n++;
		}
	par->b = n;

	/* determine number of data values */
	par->m = 3*par->b;

	/* close data file */
	fclose(in);

}



commentLine(line, n)
int n;
char *line;
{
   int c, i;
/*
   for(i = 0; i < n; i++) {
      c = (int) *(line + i);
      if (c ==  ' ' || c == '\t') continue;
      else if (c ==  '#') return(1);
      else if (c == '\n') return(1);
      else if (c == '\0') return(1);
      return(0);
      }
   return(1);
*/
}




/******************************************************************************
	readFault: read fault information file

	usage: readFault(fault, par)

	fault - fault information
	par    - parameters
******************************************************************************/
readFault(fault, par)
parameters *par;
faultType  *fault;
{
	int e, n;
	char line[82];
	FILE *in;

	/* initialize */
	n = 0;

	/* open fault file  */
	in = fopen(par->faultFile, "r");
   if (in == NULL)
      {
      fprintf(stderr, "ERROR: can't open fault file %s\n", par->faultFile);
      exit(-1);
      }

	/* read file line by line, skipping comments */
	while(fgets(line, 82, in))
		{
		if (faultComment(line)) continue;
		fault[n].ft = *line;
		e = sscanf(line + 1, "%lf %lf %lf %lf %lf %lf %lf %lf %s %d %d", 
		           &fault[n].ds,   &fault[n].dl,   &fault[n].dt,   &fault[n].dw, 
		           &fault[n].dp,   &fault[n].xLon, &fault[n].yLat, &fault[n].az,
		            fault[n].name, &fault[n].nx,   &fault[n].ny);
		if (e == 9)
			{
			fault[n].nx = 1;
			fault[n].ny = 1;
			} 
		else if (e != 11)
			{
			fprintf(stderr, "ERROR: bad fault file format\n");
			fprintf(stderr, "%s", line);
			exit(-1);
			}
		n++;
		}
	par->f = n;

	/* determine number of subfaults (fault parameters) */
	par->n = 0;
	for(n = 0; n < par->f; n++) par->n += fault[n].nx*fault[n].ny;

	/* close fault file */
	fclose(in);

}



faultComment(line)
char *line;
{
   int c;
   c = (int) *line;
	if (c == 's' || c == 'r' || c == 'l') return(0);
	if (c == 'd' || c == 'n' || c == 't') return(0);
	if (c == 'S' || c == 'R' || c == 'L') return(0);
	if (c == 'D' || c == 'N' || c == 'T') return(0);
	return(1);
}



/******************************************************************************
	getASlip: add normalized a priori slip errors

	usage: getASlip(A, B, par, faults)

	par    - parameters
	A      - parameter matrix
	B      - RHS data vector
	faults - fault information
******************************************************************************/
getASlip(A, B, par, faults)
parameters *par;
faultType  *faults;
double     **A, *B;
{
	double aSlip, aSlipError, aWeight;
	int i, j, m, n, sub;
	char line[82];
	FILE *in;

	/* initialize */
	m = par->m;
	n = 0;

	/* zero parameter matrix and data vector */
	for(i = m; i < m + par->n; i++)
		{
		for(j = 0; j < par->n; j++) A[i][j] = 0.;
		B[i] = 0.;
		}

	/* read slip file (if specified) */
	if (par->slipFile[0])
		{
		in = fopen(par->slipFile, "r");
		if (in == NULL)
			{
			fprintf(stderr, "ERROR: slip file %s not found\n", par->slipFile);
			exit(-1);
			}
		while(fgets(line, 82, in)) 
			{
			sscanf(line, "%lf", &aSlip, &aSlipError);
			if (aSlipError > 0.)
				{
				A[m][n] =    1./aSlipError;
				B[m]    = aSlip/aSlipError;
				}
			m++;
			n++;
			}
		if (n < par->n) fprintf(stderr, "WARNING: slips less than faults\n");
		fclose(in);
		}

	/* else specify uniform a priori slip (for each fault) */
	else
		{
		n = 0;
		for(i = 0; i < par->f; i++)
			{
			aSlip = faults[i].ds;
			sub   = faults[i].nx*faults[i].ny;
			aWeight = 1./par->aError;
			for(j = m; j < m + sub; j++)
				{
				A[j][n++] =    1.*aWeight;
				B[j]      = aSlip*aWeight;
				}
			m += sub;
			}
		}

	/* return a priori slips */
	return(par->n);

}




/******************************************************************************
	getSmoothness: get Laplacian 2D smoothness constraints 

	usage: getSmoothness(A, B, par, faults)

	A      - parameter matrix
	B      - RHS data vector
	par    - parameters
	faults - fault information

	NOTE: there is NO dx, dy dimension calculation
******************************************************************************/
getSmoothness(A, B, par, faults)
parameters *par;
faultType  *faults;
double     **A, *B;
{
	Laplacian smooth;
	double sWeight;
	int i, j, k, m, n, nx, ny;

	/* initialize */
	sWeight =          1./par->sError;
	m = par->m;
	n = 0;

	/* null parameter matrix (data vector nulled below) */
	for(i = m; i < m + par->n; i++) for(j = 0; j < par->n; j++) A[i][j] = 0.;
	
	/* loop over each fault and calculate smoothness for each sub-fault */
	for(i = 0; i < par->f; i++)
		{
		nx = faults[i].nx;
		ny = faults[i].ny;
		for(j = 0; j < ny; j++) 
			{
			for(k = 0; k < nx; k++)
				{
				smooth.central = 0;
				smooth.upper = smooth.lower = 0;
				smooth.rside = smooth.lside = 0;
				if (j) {smooth.upper++; smooth.central--;}
				if (k) {smooth.lside++; smooth.central--;}
				if (ny - j - 1 && ny - 1) {smooth.lower++; smooth.central--;}
				if (nx - k - 1 && nx - 1) {smooth.rside++; smooth.central--;}
				smooth.offset = nx;
/* * * * * * * * * 
cheater
*/
smooth.central = -4;

				A[m][n] += (double) smooth.central * sWeight;
				if (smooth.upper) A[m][n - smooth.offset] += sWeight;
				if (smooth.lower) A[m][n + smooth.offset] += sWeight;
				if (smooth.lside) A[m][n -             1] += sWeight;
				if (smooth.rside) A[m][n +             1] += sWeight;
				B[m] = 0.;
				m++;
				n++;
				}
			}
		}

	/* return number of smoothness contraints */
	return(par->n);

}



/******************************************************************************
	getMoment(A, B, par, faults)

	usage: getMoment(A, B, par, faults)

	A      - parameter matrix
	B      - RHS data vector
	par    - parameters
	faults - fault information
******************************************************************************/
getMoment(A, B, par, faults)
parameters *par;
faultType  *faults;
double     **A, *B;
{
	double dWeight, moment;
	int i, j, m, n, sub;

	/* initialize */
	m = par->m;
	n = 0;

	/* null parameter matrix and data vector */
	for(i = m; i < m + par->f; i++)
		{
		for(j = 0; j < par->n; j++) A[i][j] = 0.;
		B[i] = 0.;
		}

	/* add average moment constraint */
	for(i = 0; i < par->f; i++)
		{
		moment = faults[i].ds;
		sub    = faults[i].nx*faults[i].ny;
		if (par->mError[i] > 0.)
			{
			dWeight = 1. / ((double) (sub) * par->mError[i]);
			for(j = n; j < n + sub; j++) A[m][j] += dWeight;
			B[m] = moment*dWeight;
			}
		m++;
		n += sub;
		}

	/* return number of moment constraints */
	return(par->f);

}



/******************************************************************************
	getKernal: get kernal functions for dislocations

	usage: getKernal(A, par, bm, faults)

	A      - kernal dislocation matrix
	par    - parameters
	bm     - benchmark information
	faults - fault parameters
******************************************************************************/
getKernal(A, par, bm, faults)
parameters *par;
benchMark  *bm;
faultType  *faults;
double     **A;
{
	faultType fault;
	double x, y;
	int i, j, k, XY, n;

	/* initialize */
	XY = par->XY;

	/* loop over benchmarks and faults */
	for(i = 0; i < par->b; i++)
		{
		n = 0;
		for(j = 0; j < par->f; j++)
			{
			getSubFault(&fault, faults + j, -1, XY);
			for(k = 0; k < fault.nx*fault.ny; k++)
				{
				getSubFault(&fault, faults + j, k, XY);
				getFaultCoord(&x, &y, bm + i, &fault, XY);
				getDisl(&A[3*i][n], &A[3*i + 1][n], &A[3*i + 2][n], x, y, &fault);
				n++;
				}
			}
		}

}



getSubFault(fout, fin, sub, XY)
faultType *fout, *fin;
int sub, XY;
{
	double dx, dy;
	int ix, iy;

	/* get general subfault information */
	if (sub == -1)
		{
		fout->az   = fin->az;
		fout->dp   = fin->dp;
		fout->ft   = fin->ft;
		fout->ds   = fin->ds;
		fout->nx   = fin->nx;
		fout->ny   = fin->ny;
		fout->dl   = fin->dl/ (double) fin->nx;
		fout->dw   = fin->dw/ (double) fin->ny;
		fout->zEle = fin->zEle;
		strcpy(fout->name, fin->name);
		fout->yLat  = fin->yLat;
		fout->xLon  = fin->xLon;
		fout->dt   = fin->dt;
		}

	/* or get particular subfault information */
	else if (fin->nx + fin->ny - 2)
		{
		ix = sub%fin->nx;
		iy = sub/fin->nx;
		fout->dt = fin->dt + (double) iy * fout->dw*sin(fout->dp/(RAD));
		dx = fout->dl*cos(fout->az/(RAD));
		dy = fout->dl*sin(fout->az/(RAD));
		if (!XY)
			{
			dx = dx/(KMD)/cos(fin->yLat/(RAD));
			dy = dy/(KMD);
			}
		fout->xLon = fin->xLon - (double) fin->nx*dx/2. + ((double) ix + 0.5)*dx;
		fout->yLat = fin->yLat - (double) fin->nx*dy/2. + ((double) ix + 0.5)*dy;
		}

}



getFaultCoord(x, y, bm, fault, XY)
benchMark  *bm;
faultType  *fault;
double *x, *y;
int XY;
{
	double cs, sn, dx, dy;
	cs = cos(fault->az/(RAD));
	sn = sin(fault->az/(RAD));
	dx = bm->xLon - fault->xLon;
	dy = bm->yLat - fault->yLat;
	if (!XY)
		{
		dx = dx*(KMD)*cos(fault->yLat/(RAD));
		dy = dy*(KMD);
		}
	*x = cs*dx + sn*dy;
	*y = cs*dy - sn*dx;
}



getDisl(ux, uy, uz, x, y, f)
faultType *f;
double *ux, *uy, *uz, x, y;
{
	double fx, fy, z, cs, sn;

	/* get displacements at surface (depth = 0) */
	z = 0.;

	/* get displacements in fault coordinate system */
	getdisl(x, y, z, f->dl, f->dt, f->dw, f->dp, f->ft, &fx, &fy, uz);

	/* rotate displacements from fault to Cartesian coordinates */
	cs = cos(f->az/(RAD));
	sn = sin(f->az/(RAD));
	*ux = fx*cs - fy*sn;
	*uy = fx*sn + fy*cs;
	
}



/******************************************************************************
	getData: add geodetic data into RHS vector

	usage: getData(B, par, bm)

	B   - RHS data vector
	par - parameters
	bm  - benchmark information
******************************************************************************/
getData(B, par, bm)
parameters *par;
benchMark  *bm;
double     *B;
{
	int i, m;
	m = 0;
	for(i = 0; i < par->b; i++)
		{
		B[m    ] = bm[i].xDat; 
		B[m + 1] = bm[i].yDat; 
		B[m + 2] = bm[i].zDat; 
		m += 3;
		}
}



/******************************************************************************
	getUnitError: normalize kernal matrix and data vector by data errors

	usage: getUnitError(A, B, par, bm)

	A          - kernal matrix
	B          - data vector
	par        - parameters
	bm         - benchmark information (including errors)
******************************************************************************/
getUnitError(A, B, par, bm)
parameters *par;
benchMark  *bm;
double **A, *B;
{
	double sWeight, xWeight, yWeight, zWeight;
	int i, j, m;

	/* normalize dislocation kernals and geodetic data */
	m = 0;
	for(i = 0; i < par->b; i++)
		{
		xWeight = 1./bm[i].xErr;
		yWeight = 1./bm[i].yErr;
		zWeight = 1./bm[i].zErr;
		for(j = 0; j < par->n; j++)
			{
			A[m    ][j] *= xWeight;
			A[m + 1][j] *= yWeight;
			A[m + 2][j] *= zWeight;
			}
		B[m    ] *= xWeight;
		B[m + 1] *= yWeight;
		B[m + 2] *= zWeight;
		m += 3;
		}

}



/******************************************************************************
	getCov: invert kernal function to get covariance of fault slip ([At*A]-1)

	usage: getCov(cov, A, par)

	cov         - covariance of fault slip parameters
	A           - kernal function matrix
	par         - parameters
******************************************************************************/
getCov(C, A, par)
parameters *par;
double **C, **A;
{
	int i, j, k, n;

	/* initialize */
	n = par->n;

	/* compute (A transpose)*(A) */
	for(i = 0; i < n; i++)
		for(j = 0; j < n; j++)
			{
			C[i][j] = 0.;
			for(k = 0; k < par->m; k++) C[i][j] += A[k][i]*A[k][j];
			}

	/* get inverse of (A transpose)*(A) (inverse is the covariance matrix) */
	gaussj(C, n);

}



getSmoothDot(smooth, c1, c2, n)
Laplacian *smooth;
int c1, c2, n;
{
	int v1, v2;
	if      (c1 < n - smooth->offset || c1 > n + smooth->offset) 
		v1 = 0;
	else if (c1 == n)
		v1 = smooth->central;
	else if (c1 == n - smooth->offset)
		v1 = smooth->upper;
	else if (c1 == n - 1)
		v1 = smooth->lside;
	else if (c1 == n + smooth->offset)
		v1 = smooth->lower;
	else if (c1 == n + 1)
		v1 = smooth->rside;
	else 
		v1 = 0;
	if      (c2 < n - smooth->offset || c2 > n + smooth->offset) 
		v2 = 0;
	else if (c2 == n)
		v2 = smooth->central;
	else if (c2 == n - smooth->offset)
		v2 = smooth->upper;
	else if (c2 == n - 1)
		v2 = smooth->lside;
	else if (c2 == n + smooth->offset)
		v2 = smooth->lower;
	else if (c2 == n + 1)
		v2 = smooth->rside;
	else
		v2 = 0;
	return(v1*v2);
}



/******************************************************************************
	getSlip: calculate slip distribution on sub-faults

	usage: getSlip(Slip, par, A, B, C)

	Slip   - slip distribution on sub-faults
	par    - parameters
	A      - dislocation kernal normalized by error
	B      - data vector normalized by error
	C      - covariance matrix
******************************************************************************/
getSlip(Slip, par, A, B, C)
parameters *par;
double     *Slip, **A, *B, **C;
{
	double *t;
	int i, j;

	/* get temporary memory */
	t = (double *) malloc(par->n*sizeof(double));

	/* mulitply At by data vector */
	for(i = 0; i < par->n; i++)
		{
		t[i] = 0;
		for(j = 0; j < par->m; j++) t[i] += A[j][i]*B[j];
		}

	/* multiply C by (At*data) to get sub-fault slip */
	for(i = 0; i < par->n; i++) 
		{
		Slip[i] = 0.;
		for(j = 0; j < par->n; j++) Slip[i] += C[i][j]*t[j];
		}

	/* free temporary memory */
	free(t);

}




/******************************************************************************
	writeOutput: write results to standard output

	usage: writeOutput(par, bm, faults, C, Slip)

	par    - parameters
	bm     - benchmark information
	faults - fault information
	C      - covariance matrix
	Slip   - sub fault slip distribution
******************************************************************************/
writeOutput(par, bm, faults, C, Slip)
parameters *par;
benchMark  *bm;
faultType  *faults;
double     **C, *Slip;
{
	double sigma;
	int i, j, n, nx, ny, sub, sum;
	sum = 0;
	for(i = 0; i < par->f; i++)
		{
		nx  = faults[i].nx;
		ny  = faults[i].ny;
		sub = nx*ny;
		for(j = 0; j < sub; j++)
			{
			n = j%nx + nx*(ny - j/nx - 1) + sum;
			sigma = sqrt(C[n][n]);
			printf("%9.2lf %8.2lf\n", Slip[n], sigma);
			}
		sum += sub; 
		}
}
