/******************************************************************************
	readline: routines to read input file lines and return information

	fileLink - store file linklist   -> linkedList = fileLink(&list,n, file)
	initLink - first link of list    -> linkedList = initLink(list)
	lastLink - last  link of list    -> linkedList = lastLink(list)
	copyLink - copy  link to link    -> linkedList = copyLink(link)
	openList - insert list2 in list1 -> linkedList = openList(list1, list2)
	takeList - reorder list by type  -> linkedList = takeList(list, type)
	takeLink - remove link from list -> linkedList = takeLink(link)
	freeList - free list memory      -> linkedList = freeList(list)
	freeLink - free link memory      -> linkedList = freeLink(link)

	Predefined Line Types and Values (readLine and getLine):

		LINE_FOUND    1   non-null line found with readLine
		FILE_END      0   EOF encountered with readLine
		LINE_END     -1   EOF encountered on continuation line (\) with readLine

		LINE_COMMENT -2   hard comment (# as first character) (readLine/getLine)

		LINE_NULL    -3   null line found in getLine ('\0' as first character)
		LINE_BLANK   -4   blank line  (all ' ' or '\t' before '\n')
		LINE_VALUE   -5   numeric value for first character (0-9, +, -, .)
		LINE_PAR     -6   parameter line found in getLine (parameter=value)
		LINE_UNKNOWN -9   unknown line type

	author: shawn larsen
	date:   october 1992
	where:  llnl
******************************************************************************/


#include <stdio.h>
#include <string.h>
#include "readline.h"


/******************************************************************************
	fileLink: read file and store lines as linked list

	usage: linkedList = fileLink(&last, n, file)

	linkedList - linked-list of lines (first element)

	last will point to last element created, if &last == NULL, output = NULL
	n is maximum number to store (if negitive it will store all)
	hard comment lines are skipped (# as first non-white character)
	/ as last character specifies continuation line - ignored for hard comment
	memory is allocated for all linked elements
******************************************************************************/
LinkList *
fileLink(last, n, file)
LinkList **last;
int      n;
FILE *file;
{
	LinkList *next, *prev, *init;
	char line[STRING_SIZE];
	next = NULL;
	prev = NULL;
	init = NULL;
	while(n-- && readLine(line, STRING_SIZE, file)) {
		next = (LinkList *) malloc(sizeof(LinkList));
		next->prev = prev;
		next->next = NULL;
		next->line = strdup(line);
		next->type = getLine(line);
		if ( prev) prev->next = next;
		if (!prev) init = next;
		prev = next;
		} 
	if (last) *last = next; 
	return(init);
}



/******************************************************************************
	initLink: find initial line in file

	usage: linkedList = initLink(list)

	list - pointer to any element in list

	linkedList: pointer to first link in list
******************************************************************************/
LinkList *
initLink(list)
LinkList *list;
{
	if (list != NULL) while(list->prev != NULL) list = list->prev;
	return(list);
}



/******************************************************************************
	lastLink: find last line in file

	usage: linkedList = lastLink(list)

	list - pointer to any element in list

	linkedList: pointer to last link in list
******************************************************************************/
LinkList *
lastLink(list)
LinkList *list;
{
	if (list != NULL) while(list->next != NULL) list = list->next;
	return(list);
}



/******************************************************************************
	copyLink: copy a list element (allocate memory)

	usage: linkedList = copyLink(list)

	list - pointer to element in list to be copied

	linkedList: pointer to copied element
******************************************************************************/
LinkList *
copyLink(link)
LinkList *link;
{
	LinkList *copy = NULL;
	if (link) {
		copy = (LinkList *) malloc(sizeof(LinkList));
		copy->next = link->next;
		copy->prev = link->prev;
		copy->type = link->type;
		copy->line = strdup(link->line);
		}
	return(copy);
}



/******************************************************************************
	openList: insert list2 after list1 element and return element after list2

	usage: linkedList = openList(list1, list2)

	list1 - pointer to any element in list1 (list2 inserted after list1) 
	list2 - pointer to any element in list2 (should be first)

	linkedList: pointer to first element after list2 in joined list
******************************************************************************/
LinkList *
openList(list1, list2)
LinkList *list1, *list2;
{
	LinkList *last, *next;
	next = list2;
	if (list1 != NULL) {
		next = list1->next;
		if (list2 != NULL) {
			list1->next = list2;
			list2->prev = list1;
			last = lastLink(list2);
			last->next = next;
			next->prev = last;
			}
		}
	return(next);
}



/******************************************************************************
	takeList: reorder list (STARTING AT LIST) with elements of type type first

	usage: linkedList = takeList(list, type)

	list - pointer to starting reorder element in list
	type - type of link element to select

	linkedList: pointer to first link after all type elements in list

	note: elements before list are not reordered
	note: return NULL if all elements have been reordered (end of list)
******************************************************************************/
LinkList *
takeList(list, type)
LinkList *list;
int type;
{
	LinkList *takeLink(), *link, *next, *take, *init = NULL;
	take = (list) ? (list->prev) : (NULL);
	while(list) {
		next = list->next;
		if (list->type == type) {
			link = takeLink(list);
			link->next = NULL;
			link->prev = take;
			if (take) take->next = link;
			if (next) if (next->prev == take) next->prev = link;
			take = link;
			}
		else {
			if (init == NULL) init = list;
			}
		list = next;
		}
	if (init) init->prev = take;
	if (take) take->next = init;
	return(init);
}



/******************************************************************************
	takeLink: remove link from list (connect adjoining links)

	usage: linkedList = takeLink(list)

	list - pointer to element to remove

	linkedList: pointer to removed link
******************************************************************************/
LinkList *
takeLink(list)
LinkList *list;
{
	if (list != NULL) {
		if (list->prev != NULL) list->prev->next = list->next;
		if (list->next != NULL) list->next->prev = list->prev;
		}
	return(list);
}



/******************************************************************************
	freeList: free memory for all elements at and following list element 

	usage: freeList(list)

	list - pointer to first element in list to deallocate memory 
******************************************************************************/
freeList(list)
LinkList *list;
{
	if (list != NULL) {
		if (list->prev) list->prev->next = NULL;
		while(list->next != NULL) {
			list = list->next;
			free(list->prev->line);
			free(list->prev);
			}
		free(list->line);
		free(list);
		}
}



/******************************************************************************
	freeLink: free memory for linked element pointed to by list

	usage: freeLink(list)

	list - pointer to element to free memory
******************************************************************************/
freeLink(list)
LinkList *list;
{
	if (list != NULL) {
		free(list->line);
		free(list);
		}
}
