/****************************************************************************
This file contains parameters for dislocation modeling.

author: shawn larsen
date:   january 1989
where:  caltech
****************************************************************************/



/*****************
GENERAL PARAMETERS
*****************/

#define  PI     3.14159265358979    /* pi                     */
#define  PI12   0.026525823849      /* (1./(12*pi))           */
#define  RAD    57.295779513        /* 180/pi                 */
#define  KMD    111.31953           /* km per degree latitude */
#define  CMKM   1.e5                /* cm per km              */
#define  MICRO  1.e-6               /* micro unit (1x10**-6)  */



/*****************
DEFAULT PARAMETERS
*****************/

#define FILELINES  25000             /* initial allocation for input file */
#define AERROR     100.0             /* default a priori slip error       */
#define MERROR     100.0             /* default average moment error      */
#define SERROR     100.0             /* default smoothness error          */
#define XERROR       1.0             /* default x data error              */
#define YERROR       1.0             /* default y data error              */
#define ZERROR       1.0             /* default z data error              */
#define DERROR       1.0             /* default d data error              */
#define GFACTOR      1.0             /* factor for data errors            */
#define SLIPFLAG       0             /* default flag for a priori slip    */
#define SMOOTHFLAG     0             /* default flag for smoothness       */
#define MOMENTFLAG     0             /* default flag for average moment   */
#define XYFLAG         0             /* default flag for xy coordinates   */
#define STRESSFLAG     1             /* default flag for stress output    */
#define MDIGIT         2             /* default output model digits       */
#define RESFLAG        0             /* default flag for residual output  */
#define INCREMENT  0.001             /* default strain increment (km)     */
#define FRICTION   0.750             /* default internal friction coeff   */
#define LAMDA      2.8e5             /* default lame' parameters (bars)   */
#define MU         2.8e5             /* default rigidity (bars)           */
#define TIMESTEPS  10000             /* default number of time steps      */
#define TINCREMENT   1.0             /* default time increment (years)    */
#define SSTRENGTH   50.0             /* default static  shear  (bars)     */
#define DSTRENGTH   30.0             /* default dynamic shear (bars)      */
#define ASTRESS     25.0             /* default arrest  stress (bars)     */



/****************
STRESS PARAMETERS
****************/

#define FAULTSTRESS     2            /* get stresses on fault plane       */



/********************
LINE INPUT PARAMETERS
********************/
#define LINEFAULT    1                  /* fault     input line             */ 
#define LINEBM       2                  /* benchmark input line             */
#define LINECONTOUR  3                  /* contour generation line          */
#define LINEPROFILE  4                  /* profile generation line          */
#define LINEPLANE    5                  /* plane   generation line          */
#define LINEBLOCK    6                  /* block   generation line          */
#define LINEPOINT    7                  /* point   generation line          */
#define LINEDATUM    8                  /* datum   generation line          */
#define LINEOPTION   9                  /* additional options               */
#define LINELENGTH  10                  /* line-length input line           */
#define LINESTOP    20                  /* stop benchmark or fault input    */
#define LINECOMMENT -2                  /* hard comment line (#)            */
#define LINENULL    -3                  /* null line                        */
#define LINEBLANK   -4                  /* blank line                       */
#define LINEVALUE   -5                  /* line starting with number        */
#define LINEPAR     -6                  /* parameter line (parameter=value) */
#define LINEUNKNOWN -9                  /* unknown line type                */
#define LINEDATA     LINEVALUE          /* successful (data) input          */





/***************
FAULT STRUCTURE
***************/

struct fault_type {
			double     xLon;  /* long. at top (surface proj.) strike cent.    */
			double     yLat;  /* lat.  at top (surface proj.) strike cent.    */
			double     zEle;  /* elev. at top (surface proj.) strike cnet.    */
			double       dl;  /* length of fault (total)                      */
			double       dt;  /* depth to top of fault (in z direction)       */
			double       dw;  /* down dip width of fault                      */
			double       az;  /* azimuth (counter-clockwise from east)        */
			double       dp;  /* dip of fault (deg) (dips to north if az=0)   */
			double       ds;  /* dislocation magnitude                        */
			char         ft;  /* dislocation type (strike: s,r,l; dip: d,n,t) */ 
			int          nx;  /* number of fault discretizations along strike */
			int          ny;  /* number of fault discretizations along dip    */
			char   name[20];  /* fault name                                   */
			};



/******************
BENCHMARK STRUCTURE
******************/

struct bench_mark {
			char   name[20];    /* name of benchmark           */
			double xLon;        /* longitude or x of benchmark */
			double yLat;        /* latidute  or y of benchmark */
			double zEle;        /* elevation or z of benchmark */
			double xDat;        /* observed x movement         */
			double yDat;        /* observed y movement         */
			double zDat;        /* observed z movement         */
			double xMod;        /* modeled  x movement         */
			double yMod;        /* modeled  y movement         */
			double zMod;        /* modeled  z movement         */
			double xErr;        /* error in x data             */
			double yErr;        /* error in y data             */
			double zErr;        /* error in z data             */
			};



/************************
BENCHMARK GROUP STRUCTURE
************************/

struct bench_mark_group {
			char        line[256];   /* input line identifier         */
			char         name[20];   /* name of benchmark group       */
			int                 n;   /* number of benchmarks in group */
			struct bench_mark *bm;   /* benchmarks                    */
			};



/********************
LINE-LENGTH STRUCTURE
********************/

struct line_length {
			char    name [20];   /* name of line-length         */
			char    name1[20];   /* name of 1st station         */
			double      xLon1;   /* longitude or x of benchmark */
			double      yLat1;   /* latidute  or y of benchmark */
			double      zEle1;   /* elevation or z of benchmark */
			char    name2[20];   /* name of 2nd station         */
			double      xLon2;   /* longitude or x of benchmark */
			double      yLat2;   /* latidute  or y of benchmark */
			double      zEle2;   /* elevation or z of benchmark */
			double       dDat;   /* line-length change          */
			double       dMod;   /* modeled line-length change  */
			double       dErr;   /* error in line-length change */
			};



/**************************
LINE-LENGTH GROUP STRUCTURE
**************************/

struct line_length_group {
			char             line[256];   /* input line identifier           */
			char              name[20];   /* name of line-length group       */
			int                      n;   /* number of line-lengths in group */
			struct line_length     *ll;   /* line-lengths                    */
			};



/**********************
PROFILE GROUP STRUCTURE
**********************/

struct profile_group {
			char             line[256];   /* input line identifier           */
			char              name[20];   /* name of profile group           */
			int                      n;   /* number of benchmarks in group   */
			struct bench_mark      *bm;   /* benchmarks                      */
			double                  x1;   /* starting x coordinate           */
			double                  x2;   /* ending   x coordinate           */
			double                  y1;   /* starting y coordinate           */
			double                  y2;   /* ending   y coordinate           */
			};



/**********************
CONTOUR GROUP STRUCTURE
**********************/

struct contour_group {
			char             line[256];   /* input line identifier           */
			char              name[20];   /* name of contour group           */
			int                      n;   /* number of benchmarks in group   */
			struct bench_mark      *bm;   /* benchmarks                      */
			double                  x1;   /* starting x coordinate           */
			double                  x2;   /* ending   x coordinate           */
			double                  y1;   /* starting y coordinate           */
			double                  y2;   /* ending   y coordinate           */
			int                     nx;   /* number of points in x direction */
			int                     ny;   /* number of points in y direction */
			};



/*******************
STRESS/STRAIN TENSOR
********************/

struct tensor
			{
			char   name[20];    /* name of benchmark                        */
			double     xLon;    /* longitude or x of point                  */
			double     yLat;    /* latidute  or y of point                  */
			double     zEle;    /* elevation or z of point (abs(zEle) used) */
			double     Txx;     /* Txx stress                               */
			double     Tyy;     /* Tyy stress                               */
			double     Tzz;     /* Tzz stress                               */
			double     Txy;     /* Txy stress                               */
			double     Txz;     /* Txz stress                               */
			double     Tyz;     /* Tyz stress                               */
			};



/***********************
POINT GROUPING STRUCTURE
***********************/

struct point_group
			{
			char          name[20];       /* name of group                   */
			char             *line;       /* line string                     */
			int               type;       /* group type                      */
			double              x1;       /* starting x coordinate           */
			double              x2;       /* ending   x coordiante           */
			double              y1;       /* starting y coordinate           */
			double              y2;       /* ending   y coordinate           */
			double              z1;       /* starting z coordinate           */
			double              z2;       /* ending   z coordinate           */
			double              dl;       /* length of group (km)            */
			double              dw;       /* width  of group (km)            */
			double              az;       /* azimuth of group (cc from east) */
			double              dp;       /* dip of group (north for 0 az)   */
			int                 nx;       /* number of points in x direction */
			int                 ny;       /* number of points in y direction */
			int                 nz;       /* number of points in z direction */
			double           angle;       /* stress orientation azimuth      */
			double             dip;       /* stress orientation dip          */
			struct tensor  *tensor;       /* 3D tensor (stress/strain)       */
			};



/*****************************
LAPLACIAN SMOOTHNESS STRUCTURE
*****************************/

struct laplacian_2d
			{
			int  central;      /* central    value */
			int  upper;        /* upper      value */  
			int  lower;        /* lower      value */
			int  rside;        /* right side value */
			int  lside;        /* left  side value */
			int  offset;       /* offset between (upper, lower) and central */
			};



/******************************
DISLOCATION PARAMETER STRUCTURE
******************************/

struct parameter_dislocation {
			int f;              /* number of faults                            */ 
			int b;              /* number of benchmark groups                  */
			int l;              /* number of line-length groups                */
			int c;              /* number of contour groups                    */
			int p;              /* number of profile groups                    */
			int n;              /* number of lines in input file               */
			int XY;             /* flag for xy input (as opposed to lat, lon)  */
			int mDigit;         /* number of output model digits rside decimal */
			int residual;       /* flag to print residuals instead of model    */
			int seed;           /* seed for random number generator            */
			float xerror;       /* default x error for gdisl (east)            */
			float yerror;       /* default y error for gdisl (north)           */
			float zerror;       /* default z error for gdisl (vertical)        */
			float derror;       /* default d error for gdisl (length)          */
			float gfactor;      /* factor for data errors (gdisl)              */
			char      file[40]; /* name of input file (unused)                 */
         char  dataFile[40]; /* name of  data file (unused)                 */
         char faultFile[40]; /* name of fault file (unused)                 */
			};



/********************************
STRESS/STRAIN PARAMETER STRUCTURE
********************************/

struct parameter_strainstress {
			char        **file; /* lines of input files                        */
			int              b; /* number of stress/strain points              */
			int              f; /* number of faults                            */ 
			int              g; /* number of point groups                      */
			int              n; /* number of lines in input file               */
			int             XY; /* flag for xy input (as opposed to lat, lon)  */
			int         mDigit; /* number of output model digits rside decimal */
			int         stress; /* flag for stress/strain output               */
			double       lamda; /* lame' parameter (should be bars)            */
			double          mu; /* rigidity        (should be bars)            */
			double   increment; /* increment for strain (from center)          */
			double    friction; /* coefficient of internal friction            */
			char faultFile[40]; /* name of fault file (unused)                 */
			};



/********************************
FAULT/FAILURE PARAMETER STRUCTURE
********************************/

struct parameter_faultfailure 
			{
			int              f; /* number of faults                            */ 
			int              n; /* number of lines in input file               */
			int             XY; /* flag for xy input (as opposed to lat, lon)  */
			int              t; /* number of time steps                        */
			double          dt; /* time increment                              */
			double       lamda; /* lame' parameter (should be bars)            */
			double          mu; /* rigidity        (should be bars)            */
			double   increment; /* increment for strain (from center)          */
			double    friction; /* coefficient of internal friction            */
			double   dstrength; /* dynamic shear strength                      */
			double   sstrength; /* static  shear strength                      */
			double     astress; /* arrest stress                               */
			};



/*****************************
COVARIANCE PARAMETER STRUCTURE
*****************************/
 
struct parameter_gpsinv
         {
         int f;              /* number of faults                            */
         int b;              /* number of benchmarks                        */
         int m;              /* number of data points                       */
         int n;              /* number fault parameters                     */
			int a;              /* number of a priori slip subfaults           */
			int d;              /* number of moment faults (=f)                */
			int s;              /* number of smoothness constraints (=n)       */
			int XY;             /* flag for xy input (as opposed to lat, lon)  */
			double  aError;     /* a priori slip error (default) (uniform)     */
			double *mError;     /* error in average moment (equivalent cm)     */
			double  sError;     /* error in smoothness (if applied)            */
			double  xError;     /* reset default xerror                        */
			double  yError;     /* reset default yerror                        */
			double  zError;     /* reset default zerror                        */
			char  dataFile[40]; /* name of  data file                          */
			char faultFile[40]; /* name of fault file                          */
			char  slipFile[40]; /* name of  slip file (a priori slip)          */
         };



/***************************
PLOTTING PARAMETER STRUCTURE
***************************/

struct parameter_plotting
			{
			int              b; /* number of benchmarks                        */
			char  dataFile[40]; /* name of  data file                          */
			int           mode; /* plotting output mode                        */
			int      arrowmode; /* arrowmode                                   */
			double   arrowsize; /* arrowsize                                   */
			int         symbol; /* symbol type                                 */
         double  symbolsize; /* symbol size                                 */
         int     symbolfill; /* symbol size                                 */
			double       scale; /* scale factor                                */
         };



typedef struct fault_type                     faultType;
typedef struct bench_mark                     benchMark;
typedef struct bench_mark_group          benchMarkGroup;
typedef struct line_length                   lineLength;
typedef struct line_length_group        lineLengthGroup;
typedef struct profile_group               profileGroup;
typedef struct contour_group               contourGroup;
typedef struct point_group                   PointGroup;
typedef struct tensor                      strainTensor;
typedef struct tensor                      stressTensor;
typedef struct tensor                            Tensor;
typedef struct laplacian_2d                   Laplacian;
