/******************************************************************************
	getdisl: calculate displacement kernals at a point due to fault dislocation

	getdisl calculates the displacement kernals at any coordinates 
	(including depth) due to an arbitrarily oriented fault dislocation 
	in an elastic half-space.
	the displacement kernals (ux, uy, uz) are the actual displacements 
	for unit fault slip.
	the displacement kernals can be multiplied by the fault displacement
	to get the real displacement at the point (x, y, z).
   the equations used are (almost verbatum) out of manshinha and smylie,
   JGR 1967 (#18) and mansinha and smylie, BSSA 1971, (#5).
	the fault is assumed to trend east-west dipping to the north;
	the coordinates at the top of the fault as it intersects (or projects)
	to the surface centered along strike are x=0, y=0.
	this is the fault coordinate system (therefore the fault extends from
	x=-dl/2 to x=+dl/2 where dl is the length).
	the displacement is assumed to be right-lateral for strike slip faults
	and normal for dip slip faults.
	this function does not distinguish between right and left-lateral 
	dislocation or between normal and thrust dislocation.
	this is determined by the sign of the slip unit when the kernal functions
	are multiplied by the real dislocation 
	(right-lateral and normal-slip are positive).
	all coordinates must be in the same units (km are good).

	usage: getdisl(x1, x2, x3, dl, dt, dw, dp, ft, &ux, &uy, &uz)

	x1    - x coordinate of point in fault coordinate system
	x2    - y coordinate of point in fault coordinate system
	x3    - z coordinate of point in fault coordinate system (z positive down)
	dl    - length of fault (total)
	dt    - depth to top of fault (in z direction)
	dw    - down dip width of fault
	dp    - dip of fault (degrees) (it dips to north or positive y direction)
	ft    - dislocation type (s, r, l  for strike-slip; d, n, t, for dip-slip)
	ux    - x displacement at point in units of slip
	uy    - y displacement at point in units of slip
	uz    - z displacement at point in units of slip (UP IS POSITIVE)

	compile: cc -g -c getdisl.c -lm

	note: all coordinates must be in same units (km are good)
	note: the kernal functions must be multiplied by slip to get displacement.
	note: dip must be between 0 and 90.

	author: shawn larsen
	date:   january 1989   
	where:  caltech
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include "disl.h"

#define EPSILON6  1.E-6
#define EPSILON10 1.E-10

#define SQ(x) ((x)*(x))
#define CU(x) ((x)*(x)*(x))

getdisl(x1, x2, x3, dl, dt, dw, dp, ft, ux, uy, uz)
double x1, x2, x3, dl, dt, dw, dp, *ux, *uy, *uz;
char ft;
{
	double h, k;
	double asi[2], asi1[2];
	double ep;
	double cs, sn, tn, dip, r, r2, r3, q, q2, q3, si, si1, si2, si3, sign;
	double u1, u2, u3;
	int strikeSlip, dipSlip;
	int i, j;

	/* initialize */
	strikeSlip = 0;
	dipSlip    = 0;
	if (ft == (char) 's' || ft == (char) 'r' || ft == (char) 'l') strikeSlip = 1;
	if (ft == (char) 'd' || ft == (char) 'n' || ft == (char) 't') dipSlip    = 1;
	if (ft == (char) 'S' || ft == (char) 'R' || ft == (char) 'L') strikeSlip = 1;
	if (ft == (char) 'D' || ft == (char) 'N' || ft == (char) 'T') dipSlip    = 1;
	ep  = EPSILON6;
	*ux = 0.;
	*uy = 0.;
	*uz = 0.;

	/* avoid singularities by not allowing points to fall on fault plane */
	if (fabs(x2) < EPSILON6) x2 = -EPSILON6;
	if (fabs(x1) < dl/2. + EPSILON6) 
		{
		if (dp <= 89. && dp >= 1.) {
			if (fabs(atan(x3/x2) - dp/(RAD)) < EPSILON10)
				if (x3 > dt - EPSILON6 && x3 < dt + dw*sin(dp/(RAD)) + EPSILON6) 
					x2 = x2 - EPSILON6;}
		else if (dp > 89.) {
			if (fabs(x2) < EPSILON6)
				if (x3 > dt - EPSILON6 && x3 < dt + dw + EPSILON6) 
					x2 = -EPSILON6;}
		else if (dp < 1.) {
			if (fabs(atan(x3/x2) - 1./(RAD)) < EPSILON10)
				if (x3 > dt - EPSILON6 && x3 < dt + dw*sin(1./(RAD)) + EPSILON6) 
					x2 = x2 - EPSILON6;}
		}



	/* FOR INCLINED FAULTS */

	if (dp <= 89.)
		{

		dip = dp/(RAD);
		if (dp < 1.) dip = 1./(RAD);
		cs = cos(dip);
		sn = sin(dip);
		tn = sn/cs;

		/* define limits of integration -- asi = si and asi1 = si1 */
		asi[ 0] = dt/sn + dw;
		asi[ 1] = dt/sn;
		asi1[0] = dl/2.;
		asi1[1] = -dl/2.;
	
		/* define substitution variables */
		r2 =  x2*sn - x3*cs;
		r3 =  x2*cs + x3*sn;
		q2 =  x2*sn + x3*cs;
		q3 = -x2*cs + x3*sn;
	
		/* loop over limits of integration */
		for(i = 0; i < 2; i++)
			{
	
			si = asi[i];
			h = sqrt(q2*q2 + SQ(q3+si));
			si2 = si*cs;
			si3 = si*sn;
	
			for(j = 0; j < 2; j++)
				{

				si1=asi1[j];
				r = sqrt(SQ(x1 - si1) + SQ(x2 - si2) + SQ(x3 - si3));
				q = sqrt(SQ(x1 - si1) + SQ(x2 - si2) + SQ(x3 + si3));
				k = sqrt(SQ(x1 - si1) + q2*q2);
	
				/* displacement for strike-slip motion */
				if (strikeSlip)
					{
					u1 = (x1 - si1)*(2.*r2/(r*(r + r3 - si)) - (4.*q2 - 2.*x3*cs)/
				     	(q*(q + q3 + si)) - 3.*tn/(q + x3 + si3) + 4.*q2*x3*sn/CU(q) -
				     	4.*q2*q3*x3*sn*(2.*q + q3 + si)/(CU(q)*SQ(q + q3 + si))) -
				     	6.*SQ(tn)*atan(((k - q2*cs)*(q - k) + (q3 + si)*k*sn)/
				     	((x1 - si1)*(q3 + si)*cs)) +
				     	3.*atan((x1 - si1)*(r3 - si)/(r2*r)) -
				     	3.*atan((x1 - si1)*(q3 + si)/(q2*q));
					u2 = sn*(3.*tn/cs*log(q + x3 + si3) - log(r + r3 - si) -
				     	(1. + 3.*SQ(tn))*log(q + q3 + si)) + 
				     	2.*SQ(r2)*sn/(r*(r + r3 - si)) +
				     	2.*r2*cs/r - 2.*sn*(2.*x3*(q2*cs - q3*sn) + q2*(q2 + x2*sn))/
				     	(q*(q + q3 + si)) -
				     	3.*tn*(x2 - si2)/(q + x3 + si3) + 
				     	2.*(q2*cs - q3*sn - x3*SQ(sn))/q +
				     	4.*q2*x3*sn*((x2 - si2) + q3*cs)/CU(q) -
				     	4.*SQ(q2)*q3*x3*SQ(sn)*(2.*q + q3 + si)/
				     	(CU(q)*SQ(q + q3 + si));
					u3 = cs*(log(r + r3 - si) + (1. + 3.*SQ(tn))*log(q + q3 + si) -
				     	3.*tn/cs*log(q + x3 + si3)) + 
				     	2.*r2*sn/r + 2.*sn*(q2 + x2*sn)/q -
				     	2.*SQ(r2)*cs/(r*(r + r3 - si)) +
				     	(4.*q2*x3*SQ(sn) - 2.*(q2 + x2*sn)*(x3 + q3*sn))/
				     	(q*(q + q3 + si)) +
				     	4.*q2*x3*sn*((x3 + si3) - q3*sn)/CU(q) -
				     	4.*SQ(q2)*q3*x3*cs*sn*(2.*q + q3 + si)/
				     	(CU(q)*SQ(q + q3 + si));
					}
	
				/* displacement for dip-slip motion */
				if (dipSlip)
					{
					u1 = (x2 - si2)*sn*
				     	(2./r + 4./q - 4.*si3*x3/CU(q) - 3./(q + x3 + si3)) -
				     	cs*(3.*log(q + x3 + si3) + 2.*(x3 - si3)/r + 4.*(x3 - si3)/q +
				     	4.*si3*x3*(x3 + si3)/CU(q)) +
				     	3./cs*(log(q + x3 + si3) - sn*log(q + q3 + si)) +
				     	6.*x3*(cs/q - q2*sn/(q*(q + q3 + si)));
					u2 = sn*(-log(r + x1 - si1) + log(q + x1 - si1) +
				     	4.*si3*x3/(q*(q + x1 - si1)) +
				     	3.*(x1 - si1)/(q + x3 + si3) + 
				     	SQ(x2 - si2)*(2./(r*(r + x1 - si1)) +
				     	4./(q*(q + x1 - si1)) -
				     	4.*si3*x3*((2.*q + x1 - si1)/(CU(q)*SQ(q + x1 - si1))))) -
				     	cs*((x2 - si2)*(2.*(x3 - si3)/(r*(r + x1 - si1)) +
				     	4.*(x3 - si3)/(q*(q + x1 - si1)) +
				     	4.*si3*x3*(x3 + si3)*((2.*q + x1 - si1)/
				     	(CU(q)*SQ(q + x1 - si1)))) +
				     	6.*atan((x1 - si1)*(x2 - si2)/((h + x3 + si3)*(q + h))) -
				     	3.*atan((x1 - si1)*(r3 - si)/(r2*r)) +
				     	6.*atan((x1 - si1)*(q3 + si)/(q2*q))) +
				     	6.*(1./cs*atan(((k - q2*cs)*(q - k) +
				     	(q3 + si)*k*sn)/((x1 - si1)*(q3 + si)*cs)) +
				     	x3*(((SQ(sn) - SQ(cs))*(q3 + si) + 
				     	2.*q2*cs*sn)/(q*(q + x1 - si1)) +
				     	(x1 - si1)*SQ(sn)/(q*(q + q3 + si))));
					u3 = sn*((x2 - si2)*(2.*(x3 - si3)/(r*(r + x1 - si1)) +
				     	4.*(x3 - si3)/(q*(q + x1 - si1)) -
				     	4.*si3*x3*(x3 + si3)*
				     	((2.*q + x1 - si1)/(CU(q)*SQ(q + x1 - si1)))) -
				     	6.*atan((x1 - si1)*(x2 - si2)/((h + x3 + si3)*(q + h))) +
				     	3.*atan((x1 - si1)*(r3 - si)/(r2*r)) -
				     	6.*atan((x1 - si1)*(q3 + si)/(q2*q))) +
				     	cs*(log(r + x1 - si1) - log(q + x1 - si1) -
				     	2.*SQ(x3 - si3)/(r*(r + x1 - si1)) -
				     	4.*(SQ(x3 + si3) - si3*x3)/(q*(q + x1 - si1)) -
				     	4.*si3*x3*SQ(x3 + si3)*((2.*q + x1 - si1)/
				     	(CU(q)*SQ(q + x1 - si1)))) +
				     	6.*x3*(cs*sn*(2.*(q3 + si)/(q*(q + x1 - si1)) +
				     	(x1 - si1)/(q*(q + q3 + si))) -
				     	q2*(SQ(sn) - SQ(cs))/(q*(q + x1 - si1)));
					}

				sign = 1.;
				if (i + j == 1) sign = -1.;
	
				/* sum for each integration limit */
				*ux = *ux + sign*u1;
				*uy = *uy + sign*u2;
				*uz = *uz - sign*u3;
	
				/* end j loop */
				}
	 
			/* end loop over limits of integration */
			}

		/* done with inclined faults */
		}



	/* FOR VERTICAL FAULTS */

	else if (dp > 89.)
		{

		/* define limits of integration */
		asi[ 0] =  dt + dw;
		asi[ 1] =  dt;
		asi1[0] =  dl/2.;
		asi1[1] = -dl/2.;

		/* loop over limits of integration */
		for(i = 0; i < 2; i++)
			{
			si3 = asi[i];

			for(j = 0; j < 2; j++)
				{

				si1 = asi1[j];
				r = sqrt(SQ(x1 - si1) + x2*x2 + SQ(x3 - si3));
				q = sqrt(SQ(x1 - si1) + x2*x2 + SQ(x3 + si3));
	
				/* displacement for strike-slip motion */
				if (strikeSlip)
					{
					u1 = x2*(x1 - si1)*(2./(r*(r + x3 - si3)) -
					     (5.*q + 8.*si3)/(2.*q*SQ(q + x3 + si3)) +
					     4.*x3*si3*(2.*q + x3 + si3)/(CU(q)*SQ(q + x3 + si3))) +
					     3.*atan((x1 - si1)*(x3 - si3)/(r*x2)) -
					     3.*atan((x1 - si1)*(x3 + si3)/(q*x2));
					u2 = -log(r + x3 - si3) + 0.5*log(q + x3 + si3) -
					     (5.*x3 - 3.*si3)/(2.*(q + x3 + si3)) -
					     4.*x3*si3/(q*(q + x3 + si3)) +
					     SQ(x2)*(2./(r*(r + x3 - si3)) -
					     (5.*q + 8.*si3)/(2.*q*SQ(q + x3 + si3)) +
					     4.*x3*si3*(2.*q + x3 + si3)/(CU(q)*SQ(q + x3 + si3)));
					u3 = x2*(2./r - 2./q + 4.*x3*si3/CU(q) +
					     3./(q + x3 + si3) + 2.*(x3 + 3.*si3)/(q*(q + x3 + si3)));
					}
	
				/* displacement for dip-slip motion */
				if (dipSlip)
					{
					u1 = x2*(2./r + 4./q - 4.*x3*si3/CU(q) - 
					     6.*x3/(q*(q + x3 + si3)));
					u2 = -log(r + x1 - si1) + log(q + x1 - si1) +
					     (6.*SQ(x3) + 10.*x3*si3)/(q*(q + x1 - si1)) +
					     6.*x3*(x1 - si1)/(q*(q + x3 + si3)) +
					     SQ(x2)*(2./(r*(r + x1 - si1)) + 4./(q*(q + x1 - si1)) -
					     4.*x3*si3*(2.*q + x1 - si1)/(CU(q)*SQ(q + x1 - si1)));
					u3 = x2*(2.*(x3 - si3)/(r*(r + x1 - si1)) -
					     2.*(x3 + 2.*si3)/(q*(q + x1 - si1)) -
					     4.*x3*si3*(x3 + si3)*(2.*q + x1 - si1)/
					     (CU(q)*SQ(q + x1 - si1))) +
					     3.*atan((x1 - si1)*(x3 - si3)/(r*x2)) -
					     3.*atan((x1 - si1)*(x3 + si3)/(q*x2));
					}
	
				sign = 1.;
				if (i + j == 1) sign = -1.;
	
				/* sum for each integration limit */
				*ux = *ux + sign*u1;
				*uy = *uy + sign*u2;
				*uz = *uz - sign*u3;
				}
	
			/* end loop over limits of integration */
			}


		/* end of vertical faults */
		}


	/* multiply by 12 pi */
	*ux *= PI12;
	*uy *= PI12;
	*uz *= PI12;
	
	/* return */
	return;

}
