/***************************************************************************
	getexpl: calcuates deformation due to point source (gridded region)

	usage: getexpl {getpar parameters} > output_file

	compile: cc -g -o getexpl getexpl.c getpoint.c -lget -lm 

	parameters
		x0 - source X coordinate (km)           [default 0.]
		y0 - source Y coordinate (km)           [default 0.]
		z0 - source Z coordinate (km) (+z down) [default 0.]
		x1 - starting X coordindate (km)        [required]
		y1 - starting Y coordindate (km)        [required]
		dx - increment in X coordinate (km)     [required]
		dy - increment in Y coordinate (km)     [required]
		nx - number of X grid points            [required]
		ny - number of Y grid points            [required]
		z1 - depth of point (km) (+z down)      [default 0.]
		lamda - elastic constant (dyne/cm**2)   [default 3.0e11]
		mu    - elastic constant (dyne/cm**2)   [default 3.0e11]
		M - seismic moment (dyne-cm)            [required]

	input/output units 
		coordinates = km
		lamda/mu    = dyne/cm**2 (default 3.e11 dyne/cm**2)
		M           = dyne-cm
		output      = cm (+z up)
	used units
		coordinates = m
		lamda/mu    = newtons/m**2 
		M           = newtons-m
		output      = m
	conversions
		1 km          = 1000 m
		1 m           =  100 cm
		1 newton-m    = 1.0e7 dyne-cm
		1 newton/m**2 = 10 dyne/cm**2
	output format
		ux uy uz ur ut
	note: ur and ut (radial and tangential deformation) are horizontal ur and ut
	note: input and used units have z positive down 
	      output units have z positive up 
	      output units have r positive out 
	      output units have t positive counterclockwise

	author: shawn larsen
	date:   may, 2000
	where:  llnl
***************************************************************************/

#include <stdio.h>
#include <math.h>

main(int argc, char **argv) 
{
	double x0, y0, z0, x1, y1, z1, x, y, dx, dy, M, lamda, mu;
	double ux, uy, uz, ur, ut;
	int i, j, nx, ny;

	/* print usage */
	if (argc == 1) {printUsage(); exit(0);}

	/* initialize values */
	x0 = 0.;
	y0 = 0.;
	z0 = 0.;
	z1 = 0.;
	lamda = 3.e11;   /* dyne/cm**2 */
	mu    = 3.e11;   /* dyne/cm**2 */

	/* get parameters */
	setpar(argc, argv);
	getpar("x0",    "F", &x0   );
	getpar("y0",    "F", &y0   );
	getpar("z0",    "F", &z0   );
	mstpar("x1",    "F", &x1   );
	mstpar("y1",    "F", &y1   );
	mstpar("dx",    "F", &dx   );
	mstpar("dy",    "F", &dy   );
	mstpar("nx",    "d", &nx   );
	mstpar("ny",    "d", &ny   );
	getpar("z1",    "F", &z1   );
	getpar("lamda", "F", &lamda);
	getpar("mu",    "F", &mu   );
	mstpar("M",     "F", &M    );
	endpar();

	/* check values */
	if (lamda <= 0.) {
		fprintf(stderr, "Error: lamda can't be <= 0.\n");
		exit(0);
		}
	if (z1 < 0.) {
		fprintf(stderr, "Error: z1 can't be < 0.\n");
		exit(0);
		}
	if (nx <= 0 || ny <= 0) {
		fprintf(stderr, "Error: nx and ny must be > 0\n");
		exit(0);
		}

	/* convert values */
	x0    *= 1000.;
	y0    *= 1000.;
	z0    *= 1000.;
	x1    *= 1000.;
	y1    *= 1000.;
	z1    *= 1000.;
	dx    *= 1000.;
	dy    *= 1000.;
	lamda *= 1.e-5*100.*100.;
	mu    *= 1.e-5*100.*100.;
	M     *= 1.e-7;

	/* loop over all points; get and print displacements */
	y = y1;
	for(j = 0; j < ny; j++) {
		x = x1;
		for(i = 0; i < nx; i++) {
			getpoint(x0, y0, z0, x, y, z1, M, lamda, mu, &ux, &uy, &uz);
			geturut (x0, y0, z0, x, y, z1, ux, uy, uz, &ur, &ut);
			printf  ("%+7.4le %+7.4le %+7.4le ", ux*100., uy*100., uz*100.);
			printf  ("%+7.4le %+7.4le\n",        ur*100., ut*100.         );
			x += dx;
			}
		y += dy;
		}

}

printUsage()
{
	fprintf(stderr, "usage: getexpl {getpar parameters} > output_file\n");
	fprintf(stderr, "x0 - source X coordinate (km)           [default 0.]\n");
	fprintf(stderr, "y0 - source Y coordinate (km)           [default 0.]\n");
	fprintf(stderr, "z0 - source Z coordinate (km) (+z down) [default 0.]\n");
	fprintf(stderr, "x1 - starting X coordindate (km)        [required]\n");
	fprintf(stderr, "y1 - starting Y coordindate (km)        [required]\n");
	fprintf(stderr, "dx - increment in X coordindte (km)     [required]\n");
	fprintf(stderr, "dy - increment in Y coordinate (km)     [required]\n");
	fprintf(stderr, "nx - number of X grid points            [required]\n");
	fprintf(stderr, "ny - number of Y grid points            [required]\n");
	fprintf(stderr, "z1 - depth of points (km) (+z down)     [default 0.]\n");
	fprintf(stderr, "lamda - elastic constant (dyne/cm**2)   [default 3.e11]\n");
	fprintf(stderr, "mu    - elastic constant (dyne/cm**2)   [default 3.e11]\n");
	fprintf(stderr, "M - seismic moment (dyne-cm)            [required]\n");
	fprintf(stderr, "output format: ux uy uz ur ut\n");
	fprintf(stderr, "output deformation units = cm (+z up)\n");
}

/* compute radial and tangental components of deformation (horizontal) */
geturut (x0, y0, z0, x1, y1, z1, ux, uy, uz, ur, ut)
double x0, y0, z0, x1, y1, z1, ux, uy, uz, *ur, *ut;
{
	double angle, x, y, z, cs, sn;
	x = x1 - x0;
	y = y1 - y0;
	if (x == 0. && y == 0.) {
		*ur = sqrt(ux*ux + uy*uy);
		*ut = 0.;
		}
	else {
		angle = atan2(y, x);
		cs = cos(angle);
		sn = sin(angle);
		*ur = ux*cs + uy*sn;
		*ut = uy*cs - ux*sn;
		}
}

