/******************************************************************************
	getpoint: calculate displacement kernals at a point due to explosive source

	getpoint calculates the displacement kernals at any coordinates (including
	depth) due to a dilitational explosive source in an elastic half-space. mks
	units are assumed. The equations are from Okada 1992, BSSA (1018-1040). 

	usage: getpoint(x0, y0, z0, x1, y1, z1, M, lamda, mu, &ux, &uy, &uz)

	x0    - x coordinate of source (m)
	y0    - y coordinate of source (m)
	z0    - z coordinate of source (m), z positive down
	x1    - x coordinate of point  (m)  
	y1    - y coordinate of point  (m)
	z1    - z coordinate of point  (m), z positive down
	M     - moment (newton-m)
	lamda - elastic constant (newtons)
	mu    - elastic constant (newtons)
	ux    - x displacement at point (m)
	uy    - y displacement at point (m)
	uz    - z displacement at point (m), z positive up

	compile: cc -g -c getpoint.c -lm

	author: shawn larsen
	date:   april 2000
	where:  llnl
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include "disl.h"

#define SQ(x) ((x)*(x))
#define CU(x) ((x)*(x)*(x))
#define C5(x) ((x)*(x)*(x)*(x)*(x))

/*
#define PI (3.14159265358979)
*/

void
getpoint(x0, y0, z0, x1, y1, z1, M, lamda, mu, ux, uy, uz)
double x0, y0, z0, x1, y1, z1, M, lamda, mu, *ux, *uy, *uz;
{
	double c, d, f, x, y, z, R, S, R3, R5, S3, C3, sqrt(), vx, vy, vz;
	double alpha, factor;

	/* initialize and reverse coordinates as appropriate */
	x  = x1 - x0;
	y  = y1 - y0;
	z  = -z1;
	c  =  z0;
	d  = c - z;
	f  = c + z;
	R  = sqrt(SQ(x) + SQ(y) + SQ(d));
	S  = sqrt(SQ(x) + SQ(y) + SQ(f));
	R3 = CU(R);
	R5 = C5(R);
	S3 = CU(S);
	C3 = 1 - ((3*SQ(d))/(SQ(R)));
	alpha = (lamda + mu)/(lamda + 2.*mu);
	vx = 0.;
	vy = 0.;
	vz = 0.;

	/* return if unusual input */
	if (R == 0. || S == 0. || mu == 0.) {
		*ux = vx;
		*uy = vy;
		*uz = vz;
		return;
		}

	/* calculate displacements (part 1A) */
	factor = -(1 - alpha)/2.;
	vx += factor*(x/R3);
	vy += factor*(y/R3);
	vz += factor*(d/R3);

	/* calculate displacements (part 1B) */
	factor = -(1 - alpha)/2.;
	vx -= factor*(x/S3);
	vy -= factor*(y/S3);
	vz -= factor*(f/S3);

	/* calculate displacements (part 2 ) */
	factor = (1 - alpha)/alpha;
	vx += factor*(x/R3);
	vy += factor*(y/R3);
	vz += factor*(d/R3);

	/* calculate displacements (part 3 ) */
	factor = (1 - alpha);
	vx += z*factor*(3*x*d/R5);
	vy += z*factor*(3*y*d/R5);
	vz += z*factor*(C3/R3);

	/* multiple by moment and constants */
	factor = M/(2*PI*mu);
	vx *= factor;
	vy *= factor;
	vz *= factor;

	/* return values */
	*ux = vx;
	*uy = vy;
	*uz = vz;

}

